#pragma once

#ifndef __cplusplus
#error "This is a C++ header"
#endif

#include "PreCompiledHeaders.hh"
#include "VivyApplication.hh"
#include "Lib/Utils.hh"
#include "Lib/Uuid.hh"
#include "Lib/Log.hh"

namespace Vivy
{
class AbstractDocument : public QObject {
    Q_OBJECT
    VIVY_UNMOVABLE_OBJECT(AbstractDocument)

protected:
    VIVY_APP_LOGGABLE_OBJECT(AbstractDocument, logger)

    void copyOrRenameWith(const QFileInfo &newFile, auto action, auto success)
    {
        const QFileInfo oldFile(getName());

        // Create folder if needed
        QDir dirOp;
        const QString newAbsDirPath = newFile.dir().absolutePath();
        if (!dirOp.exists(newAbsDirPath)) {
            logInfo() << "Create folder " << VIVY_LOG_QUOTED(newAbsDirPath);
            dirOp.mkpath(newAbsDirPath);
        }

        if (newFile.absoluteFilePath() == oldFile.absoluteFilePath()) {
            throw std::runtime_error("Can't rename or copy a file to itself!");
        }

        if (newFile.exists()) {
            logWarning() << "Deleting the already existing file " << VIVY_LOG_QUOTED(newFile);
            if (!dirOp.remove(newFile.absoluteFilePath()))
                throw std::runtime_error("Failed to remove " +
                                         newFile.absoluteFilePath().toStdString());
        }

        if (action(oldFile.absoluteFilePath(), newFile.absoluteFilePath())) {
            success();
            save();
        }

        else {
            throw std::runtime_error("Failed to copy or rename " +
                                     oldFile.absoluteFilePath().toStdString() + " to " +
                                     newFile.absoluteFilePath().toStdString());
        }
    }

public:
    enum class Type : quint64 {
        Vivy   = Utils::toUnderlying(Utils::DocumentType::Vivy),
        Script = Utils::toUnderlying(Utils::DocumentType::VivyScript)
    };

protected:
    AbstractDocument(Type childType, const QString &documentName)
        : type(childType)
        , name(documentName)
    {
    }

    // Automate a part of the rename process, just need to provide a "success"
    // callback and the new file info.
    void copyWith(const QFileInfo &newFile, auto success)
    {
        auto action = [](const QString &oldFile, const QString &newFileName) noexcept -> bool {
            return QFile::copy(oldFile, newFileName);
        };
        copyOrRenameWith(newFile, action, success);
    }

    // Automate a part of the rename process, just need to provide a "success"
    // callback and the new file info.
    void renameWith(const QFileInfo &newFile, auto success)
    {
        auto action = [](const QString &oldFile, const QString &newFileName) noexcept -> bool {
            return QFile::rename(oldFile, newFileName);
        };
        copyOrRenameWith(newFile, action, success);
    }

    Type type;
    QString name{};
    const Uuid uuid{};

public:
    virtual void copy(const QString &)   = 0;
    virtual void rename(const QString &) = 0;
    virtual void save()                  = 0;

    QString getName() const noexcept { return name; }
    Type getType() const noexcept { return type; }
    Uuid getUuid() const noexcept { return uuid; }

signals:
    void documentChanged();
};
}

bool operator==(const Vivy::AbstractDocument &a, const Vivy::AbstractDocument &b) noexcept;
