#pragma once

#ifndef __cplusplus
#error "This is a C++ header"
#endif

extern "C" {
#include <libavutil/opt.h>
#include <libavcodec/avcodec.h>
#include <libavformat/avformat.h>
#include <libswresample/swresample.h>
#include <libavcodec/avfft.h>
}

#include "PreCompiledHeaders.hh"
#include "VivyApplication.hh"
#include "Lib/Utils.hh"
#include "Lib/Log.hh"

namespace Vivy
{
static inline constexpr auto codecContexteleter = [](AVCodecContext *ptr) noexcept -> void {
    if (ptr)
        avcodec_free_context(&ptr);
};
static constexpr inline auto dataDeleter = [](double *ptr) noexcept -> void {
    if (ptr)
        av_free(ptr);
};
static constexpr inline auto avFrameDeleter = [](AVFrame *ptr) noexcept -> void {
    if (ptr)
        av_frame_free(&ptr);
};

static constexpr inline auto swrContenxtDeleter = [](SwrContext *swr) noexcept -> void {
    if (swr)
        swr_free(&swr);
};

static inline constexpr auto avFormatContextDeleter = [](AVFormatContext *ptr) noexcept -> void {
    if (ptr)
        avformat_free_context(ptr);
};

using AVFormatContextPtr = std::unique_ptr<AVFormatContext, decltype(avFormatContextDeleter)>;
using AVCodecContextPtr  = std::unique_ptr<AVCodecContext, decltype(codecContexteleter)>;
using AVFramePtr         = std::unique_ptr<AVFrame, decltype(avFrameDeleter)>;
using SwrContextPtr      = std::unique_ptr<SwrContext, decltype(swrContenxtDeleter)>;

// Simple abstract class to be inherited by video streams and audio streams.
template <size_t AVMEDIA_TYPE> class AbstractMediaStream {
    VIVY_UNMOVABLE_OBJECT(AbstractMediaStream)

public:
    static inline constexpr AVMediaType avMediaType = static_cast<AVMediaType>(AVMEDIA_TYPE);

protected:
    VIVY_APP_LOGGABLE_OBJECT(AbstractMediaStream, logger)
    using Super = AbstractMediaStream<AVMEDIA_TYPE>;

protected:
    AbstractMediaStream(const AVCodec *streamCodec, AVFormatContext *formatArg, AVStream *streamArg,
                        int index)
        : codecId(streamArg->codecpar->codec_id)
        , codec(streamCodec)
        , codecParams(streamArg->codecpar)
        , dataFormat(formatArg)
        , stream(streamArg)
        , streamIndexInContext(index)
    {
        if (codec == nullptr)
            throw std::runtime_error("failed to find a decoder for stream");

        codecContext.reset(avcodec_alloc_context3(codec));
        if (!codecContext)
            throw std::runtime_error("failed to allocate codec context");

        if (avcodec_parameters_to_context(codecContext.get(), codecParams) < 0)
            throw std::runtime_error("failed to copy parameters to codec context");

        if (avcodec_open2(codecContext.get(), codec, nullptr) < 0)
            throw std::runtime_error("failed to open audio decoder for a stream");

        VIVY_LOG_CTOR() << "Codec: " << VIVY_LOG_QUOTED(codec->name) << ", id: " << codecId;
        VIVY_LOG_CTOR() << "Sample rate: " << codecParams->sample_rate;
        VIVY_LOG_CTOR() << "Bit rate: " << codecParams->bit_rate;
        VIVY_LOG_CTOR() << "Channels: " << codecParams->ch_layout.nb_channels;
    }

public:
    virtual ~AbstractMediaStream() noexcept {}
    int getStreamIndex() const noexcept { return streamIndexInContext; }
    QString getName() const noexcept { return QString::fromUtf8(codec->name); }
    AVCodecID getCodecId() const noexcept { return codecId; }

    virtual QJsonObject getProperties() const noexcept
    {
        QJsonObject ret;
        ret.insert("Codec name", codec->name);
        return ret;
    }

protected:
    // Codec related informations
    AVCodecID codecId{ AV_CODEC_ID_NONE };
    const AVCodec *codec{ nullptr };
    AVCodecParameters *codecParams{ nullptr };
    AVCodecContextPtr codecContext{ nullptr };
    AVFormatContext *dataFormat{ nullptr };

    // Stream is held by AudioContext
    AVStream *stream{ nullptr };

    // Store the index of this stream
    int streamIndexInContext;
};

template <typename Stream> class AbstractMediaContext {
    VIVY_UNMOVABLE_OBJECT(AbstractMediaContext)

public:
    static inline constexpr AVMediaType avMediaType = Stream::avMediaType;
    using StreamPtr                                 = std::shared_ptr<Stream>;
    using StreamWeakPtr                             = std::weak_ptr<Stream>;

protected:
    VIVY_APP_LOGGABLE_OBJECT(AbstractMediaContext, logger)
    using Super = AbstractMediaContext<Stream>;

public:
    AbstractMediaContext(const QString &path)
        : filePath(path)
    {
        if (!format)
            throw std::runtime_error("out of memory, can't create allocate the AVFormatContext");

        const std::string filePathStdHolder = filePath.toStdString();
        const char *filename                = filePathStdHolder.c_str();
        AVFormatContext *formatPtr          = format.get();

        // Get the format from the file
        if (avformat_open_input(&formatPtr, filename, nullptr, nullptr) != 0) {
            [[maybe_unused]] void *relatedOnFailure =
                format.release(); // freed by avformat_open_input
            throw std::runtime_error("failed to open file");
        }

        if (avformat_find_stream_info(formatPtr, nullptr) < 0) {
            throw std::runtime_error("failed to get audio stream info");
        }

        // Populate all the stream indexes
        for (uint i = 0; i < format->nb_streams; ++i) {
            AVStream *itFmt           = format->streams[i];
            AVCodecParameters *params = itFmt->codecpar;
            const AVCodec *streamCodec      = avcodec_find_decoder(params->codec_id);
            if (streamCodec && streamCodec->type == avMediaType)
                audioStreams.insert(i, std::make_shared<Stream>(streamCodec, formatPtr, itFmt, i));
        }

        // Get the default stream
        defaultStreamIndex = av_find_best_stream(formatPtr, avMediaType,
                                                 -1, // Let AV find one stream
                                                 -1, // We don't want related streams
                                                 nullptr, 0);
        if (defaultStreamIndex < 0)
            logError() << "Could not find the best stream";

        logDebug() << "Opened context for " << VIVY_LOG_QUOTED(path) << " with duration "
                   << formatPtr->duration << " and default stream index " << defaultStreamIndex;
    }

    virtual ~AbstractMediaContext() {}

    StreamWeakPtr getStream(int index) const noexcept
    {
        if (index < 0)
            return StreamWeakPtr{ spareNullSreamPtr };

        uint unsignedIndex = static_cast<uint>(index);
        const auto found   = audioStreams.find(unsignedIndex);

        if (found != audioStreams.end())
            return StreamWeakPtr{ *found };

        return StreamWeakPtr{ spareNullSreamPtr };
    }

    StreamWeakPtr getDefaultStream() const noexcept
    {
        return (defaultStreamIndex < 0) ? StreamWeakPtr{ spareNullSreamPtr }
                                        : getStream(defaultStreamIndex);
    }

    virtual QString getElementName() const noexcept
    {
        return "Context: " + QFileInfo(filePath).baseName();
    }

    virtual QJsonDocument getProperties() const noexcept
    {
        QJsonDocument ret;
        QJsonArray streams;

        QFileInfo file(filePath);

        for (const auto &audioStreamPtr : audioStreams) {
            streams.append(audioStreamPtr->getProperties());
        }

        QJsonObject self{ { "Streams", streams }, { "Base name", file.baseName() } };

        ret.setObject(self);
        return ret;
    }

private:
    AVFormatContextPtr format{ avformat_alloc_context(), avFormatContextDeleter };

    const QString filePath;               // Usefull information
    QMap<uint, StreamPtr> audioStreams{}; // THe audio streams of the file

    int defaultStreamIndex{ -1 };

    // Spare always null shared pointer, to be used when the audioStream[i] was
    // not found.
    StreamPtr spareNullSreamPtr{ nullptr };
};
}
