#ifndef VIVY_DOCUMENT_H
#define VIVY_DOCUMENT_H

#ifndef __cplusplus
#error "This is a C++ header"
#endif

#include "PreCompiledHeaders.hh"
#include "Lib/Utils.hh"
#include "Lib/AbstractDocument.hh"
#include "Lib/Uuid.hh"
#include "Lib/Document/CRTPSubDocument.hh"

#define DCL_VIVY_SAVE_KEY(name) static const inline QString Key##name = QStringLiteral(#name);

namespace Vivy
{
class VivyDocument final : public AbstractDocument {
    Q_OBJECT
    VIVY_UNMOVABLE_OBJECT(VivyDocument)

    DCL_VIVY_SAVE_KEY(Capabilities)
    DCL_VIVY_SAVE_KEY(Name)
    DCL_VIVY_SAVE_KEY(SubDocuments)
    DCL_VIVY_SAVE_KEY(SubAudio)
    DCL_VIVY_SAVE_KEY(SubAss)
    DCL_VIVY_SAVE_KEY(SubVideo)
    DCL_VIVY_SAVE_KEY(InternalAssSource)
    DCL_VIVY_SAVE_KEY(HaveInternalAss)
    DCL_VIVY_SAVE_KEY(DocumentVersion)

    using LoadFunctionForVersion =
        std::pair<const char *, void (*)(VivyDocument *const, const QJsonDocument &)>;

public:
    enum Capabilities : quint64 { AudioAble = (1 << 1), VideoAble = (1 << 2), AssAble = (1 << 3) };
    enum Options : quint64 {
        NoOption               = 0,
        UntouchedByDefault     = (1 << 1),
        MemoryDocumentCreation = (1 << 2)
    };

    static QString runtimeVersion;

    static inline const QString fileSuffix               = QStringLiteral("vivy");
    static inline constexpr Utils::DocumentType type     = Utils::DocumentType::Vivy;
    static inline constexpr quint64 possibleCapabilities = AudioAble | VideoAble | AssAble;

private:
    /* The document name, the filepath is {location}/{name}.{suffix} */
    QString documentName;
    QDir documentLocation{ QDir::current() };

    /* The type to get all fields, has audio, etc... */
    quint64 documentType{ 0 };
    quint64 documentOptions{ 0 };

    /* Links to other files, they're not embeded inside the vivy file */
    std::shared_ptr<AudioSubDocument> audioDocument{ nullptr };
    std::shared_ptr<VideoSubDocument> videoDocument{ nullptr };
    std::shared_ptr<AssSubDocument> assDocument{ nullptr };

    static bool detectDocumentType(const QFileInfo &file, Capabilities *const) noexcept;

    void addDocumentType(Capabilities) noexcept;
    void saveMemoryFile(const QFileInfo &);
    QFileInfo computeFileInfo(const QString &) const noexcept;

    QJsonDocument getSaveJsonDocumentFile() const;
    void loadSaveJsonDocumentFile(const QJsonDocument &);

    static void loadSaveJsonDocumentFile_ALPHA(VivyDocument *const, const QJsonDocument &);

    static constexpr inline auto loadFunctions = { LoadFunctionForVersion{
        "alpha", loadSaveJsonDocumentFile_ALPHA } };

public:
    // Create an empty document
    explicit VivyDocument(const QString &name, Options opt = NoOption);

    void copy(const QString &) override;
    void rename(const QString &) override;
    void save() override;
    bool loadSubDocument(const QString &) noexcept;
    bool loadSubDocument(const QString &, Capabilities) noexcept;
    bool loadSubDocument(const QJsonValue &, const QString &, VivyDocument::Capabilities) noexcept;

    // Getters
    std::shared_ptr<AudioSubDocument> getAudioSubDocument() const noexcept;
    std::shared_ptr<VideoSubDocument> getVideoSubDocument() const noexcept;
    std::shared_ptr<AssSubDocument> getAssSubDocument() const noexcept;

    QString getDocumentCapabilitiesString() const noexcept;

    bool checkDocumentCapabilities(Capabilities) const noexcept;
    bool checkDocumentOption(Options) const noexcept;

    // Setters
    void setAudioSubDocument(const QString) noexcept;
    void setVideoSubDocument(const QString) noexcept;
    void setAssSubDocument(const QString) noexcept;

    // PropertyConstViewable
    QString getElementName() const noexcept;
    QJsonDocument getProperties() const noexcept;
};

}

#endif // VIVY_DOCUMENT_H
