/// Enum used to represent drawing commands.
///
/// Things you should know:
/// - Commands must appear after {\p1+} and before {\p0}. (except for \clip(..))
/// - Drawings must always start with a move to command.
/// - Drawings must form a closed shape.
/// - All unclosed shape will be closed with a straight line automatically.
/// - Overlapping shapes in the Dialogue line will be XOR-ed with each-other.
/// - If the same command follows another, it isn’t needed to write its identifier letter again,
///   only the coordinates.
/// - The coordinates are relative to the current cursor position (baseline) and the alignment
///   mode.
/// - Commands p and c should only follow other b-spline commands.
///
/// Examples:
/// - Square: `m 0 0 l 100 0 100 100 0 100`
/// - Rounded square: `m 0 0 s 100 0 100 100 0 100 c`
///   (c equals to `p 0 0 100 0 100 100` in this case)
/// - Circle (almost): `m 50 0 b 100 0 100 100 50 100 b 0 100 0 0 50 0`
///   (note that the 2nd 'b' is optional here)
#[derive(Debug, Clone, PartialEq)]
pub enum ASSDrawingCmd {
    // Moves the cursor to <x>, <y>
    M {
        x: i64,
        y: i64,
    },

    // Moves the cursor to <x>, <y> (unclosed shapes will be left open)
    N {
        x: i64,
        y: i64,
    },

    // Draws a line to <x>, <y>
    L {
        x: i64,
        y: i64,
    },

    // 3rd degree bezier curve to point 3 using point 1 and 2 as the control points
    B {
        x1: i64,
        y1: i64,
        x2: i64,
        y2: i64,
        x3: i64,
        y3: i64,
    },

    // 3rd degree uniform b-spline to point N, must contain at least 3 coordinates
    S {
        x1: i64,
        y1: i64,
        x2: i64,
        y2: i64,
        x3: i64,
        y3: i64,
        others: Vec<(i64, i64)>,
    },

    // Extend b-spline to <x>, <y>
    P {
        x: i64,
        y: i64,
    },

    // close b-spline
    C,
}

/// Contains an ASS drawing.
#[derive(Debug, Default, Clone, PartialEq)]
pub struct ASSDrawing {
    content: Vec<ASSDrawingCmd>,
}

impl ASSDrawingCmd {
    pub fn is_move_cmd(&self) -> bool {
        use ASSDrawingCmd::*;
        matches!(self, M { .. } | N { .. })
    }
}

impl ASSDrawing {
    /// Verify that the drawing is correct!
    pub fn verify(&self) -> bool {
        match &self.content[..] {
            [] => true,
            [head] => head.is_move_cmd(),
            whole @ [head, content @ ..] if head.is_move_cmd() => {
                let mut in_bsplit_line = false;
                for (prev, curr) in whole.iter().zip(content) {
                    match curr {
                        ASSDrawingCmd::M { .. } | ASSDrawingCmd::N { .. } => {
                            // FIXME: This is incorrect for N, it could be at the last position,
                            //        check latter if we can find a moar correct version of the
                            //        check function...
                            log::error!(target: "ass", "found move cmd `{curr:?}` in the middle of a drawing");
                            return false;
                        }

                        ASSDrawingCmd::C | ASSDrawingCmd::P { .. } => {
                            if !matches!(prev, ASSDrawingCmd::S { .. }) {
                                log::error!(target: "ass", "found C/P cmd not after an S command");
                                return false;
                            }
                            in_bsplit_line = false;
                        }

                        ASSDrawingCmd::S { .. } => in_bsplit_line = true,
                        ASSDrawingCmd::B { .. } | ASSDrawingCmd::L { .. } => {}
                    }
                }
                !in_bsplit_line
            }
            _ => false,
        }
    }
}

impl std::fmt::Display for ASSDrawingCmd {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            ASSDrawingCmd::M { x, y } => write!(f, "m {x} {y}"),
            ASSDrawingCmd::N { x, y } => write!(f, "n {x} {y}"),
            ASSDrawingCmd::L { x, y } => write!(f, "l {x} {y}"),
            ASSDrawingCmd::B { x1, y1, x2, y2, x3, y3 } => write!(f, "b {x1} {y1} {x2} {y2} {x3} {y3}"),
            ASSDrawingCmd::S { x1, y1, x2, y2, x3, y3, others } if others.is_empty() => {
                write!(f, "b {x1} {y1} {x2} {y2} {x3} {y3}")
            }
            ASSDrawingCmd::S { x1, y1, x2, y2, x3, y3, others } => {
                let others = others
                    .iter()
                    .map(|(x, y)| format!("{x} {y}"))
                    .collect::<Vec<_>>()
                    .join(" ");
                write!(f, "b {x1} {y1} {x2} {y2} {x3} {y3} {others}")
            }
            ASSDrawingCmd::P { x, y } => write!(f, "p {x} {y}"),
            ASSDrawingCmd::C => write!(f, "c"),
        }
    }
}

impl std::fmt::Display for ASSDrawing {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let content = self
            .content
            .iter()
            .map(|cmd| format!("{cmd}"))
            .collect::<Vec<_>>()
            .join(" ");
        write!(f, "{content}")
    }
}
