use crate::{
    definitions::{ASSEvent, ASSFileSection, ScriptInfoKey},
    reader::ASSElementReader,
    ASSColor, ASSContainer, ASSElementReaderError, ASSLine, ASSStyle,
};
use std::collections::HashMap;
use vvs_procmacro::EnumVariantFromStr;

/// Line fields of the `V4+ Styles` line in the ASS file.
#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash, EnumVariantFromStr)]
enum V4PlusStyleFields {
    Name,
    Fontname,
    Fontsize,
    PrimaryColour,
    SecondaryColour,
    OutlineColour,
    BackColour,
    Bold,
    Italic,
    Underline,
    StrikeOut,
    ScaleX,
    ScaleY,
    Spacing,
    Angle,
    BorderStyle,
    Outline,
    Shadow,
    Alignment,
    MarginL,
    MarginR,
    MarginV,
    Encoding,
}

/// Line fields of the `V4+ Styles` line in the ASS file.
#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash, EnumVariantFromStr)]
enum EventFields {
    Marked,
    Layer,
    Start,
    End,
    Style,
    Name,
    MarginL,
    MarginR,
    MarginV,
    Effect,
    Text,
}

/// Helper wrapper struct to auto unwrap things when we get entries from the HashMap.
#[derive(Debug)]
struct UnwrapHashMap<'a, K>(HashMap<K, &'a str>)
where
    K: Eq + Copy + std::hash::Hash + std::fmt::Debug;

impl<'a, K: Eq + Copy + std::hash::Hash + std::fmt::Debug> UnwrapHashMap<'a, K> {
    /// Get the element from the hashmap, if the element is not present just panics.
    pub fn get(&self, key: K) -> &str {
        self.0.get(&key).unwrap_or_else(|| panic!("failed to find key {key:?}"))
    }

    /// Try to get the element from the hashmap, if the element is not present, returns the
    /// specified default value instead.
    pub fn get_or(&self, key: K, default: &'a str) -> &str {
        self.0.get(&key).copied().unwrap_or(default)
    }

    /// Try to get the element from the hashmap, if the element is present then tries to parse it.
    pub fn parsed<T>(
        &self,
        key: K,
        parser: fn(&str, name: &str) -> Result<T, ASSElementReaderError>,
        name: &str,
    ) -> Result<T, ASSElementReaderError> {
        let Self(hashmap) = self;
        let element = hashmap.get(&key).ok_or(ASSElementReaderError::Custom(format!(
            "failed to find entry with key {key:?}"
        )))?;
        parser(element, name)
    }

    /// Try to get the element from the hashmap, if the element is present then tries to parse it.
    /// If the element is not present or the parsing failed, returns the default value.
    pub fn parsed_or<T>(
        &self,
        key: K,
        parser: fn(&str, &str) -> Result<T, ASSElementReaderError>,
        name: &str,
        default: T,
    ) -> T {
        self.parsed(key, parser, name).unwrap_or(default)
    }
}

/// Documentation available here: http://www.tcax.org/docs/ass-specs.html or in the `utils/manual`
/// folder from the root of the vvs project.
#[derive(Debug, Default)]
pub struct ASSReader {
    section: Option<ASSFileSection>,

    styles_format: Vec<V4PlusStyleFields>,
    events_format: Vec<EventFields>,

    script_info: HashMap<ScriptInfoKey, String>,
    styles: HashMap<String, ASSStyle>,
    events: Vec<ASSEvent>,
}

fn parse_color(color: &str, name: &str) -> Result<ASSColor, ASSElementReaderError> {
    ASSColor::try_from_bgra(color).map_err(|err| ASSElementReaderError::Custom(format!("invalid {name} color: {err}")))
}

fn parse_boolean(boolean: &str, name: &str) -> Result<bool, ASSElementReaderError> {
    match boolean.trim() {
        "-1" | "1" | "+1" => Ok(true),
        "0" => Ok(false),
        boolean => Err(ASSElementReaderError::Custom(format!(
            "invalid ass boolean for {name} found: {boolean}"
        ))),
    }
}

fn parse_float(float: &str, name: &str) -> Result<f64, ASSElementReaderError> {
    float
        .parse::<f64>()
        .map_err(|err| ASSElementReaderError::Custom(format!("invalid float value for {name}: {err}")))
}

fn parse_int(int: &str, name: &str) -> Result<i64, ASSElementReaderError> {
    int.parse::<i64>()
        .map_err(|err| ASSElementReaderError::Custom(format!("invalid integer value for {name}: {err}")))
}

/// Parse dates in the `0:00:00:00` format
fn parse_date(date: &str, name: &str) -> Result<i64, ASSElementReaderError> {
    let std_err = "invalid date for {name}, expected \"h:mm:ss.cc\", got {date:?}".to_string();
    let [h, m, sc] = &date.split(':').collect::<Vec<_>>()[..] else {
        return Err(ASSElementReaderError::Custom(std_err));
    };
    let [s, c] = &sc.split('.').collect::<Vec<_>>()[..] else {
        return Err(ASSElementReaderError::Custom(std_err));
    };
    let check_compnent = |str: &str, compnent: &str, len: usize| {
        if str.len() > len {
            Err(ASSElementReaderError::Custom(std_err.clone()))
        } else {
            Ok(str.parse::<u16>().map_err(|err| {
                ASSElementReaderError::Custom(format!("invalid component {compnent} for date {name}: {err}"))
            })? as i64)
        }
    };
    let (h, m, s, c) = (
        check_compnent(h, "hours", 1)?,
        check_compnent(m, "minutes", 2)?,
        check_compnent(s, "seconds", 2)?,
        check_compnent(c, "centi-seconds", 2)?,
    );
    Ok(((h * 60 + m) * 60 + s) * 100 + c)
}

impl ASSReader {
    fn read_script_info(&mut self, line: &str) -> Result<(), ASSElementReaderError> {
        let Some((key, value)) = line.split_once(':') else {
            return Err(ASSElementReaderError::Custom(format!(
                "invalid script info line: {line}"
            )));
        };
        let value = value.trim();
        let key = match key
            .trim()
            .parse::<ScriptInfoKey>()
            .map_err(ASSElementReaderError::Custom)?
        {
            ScriptInfoKey::ScriptType if value.ne("V4.00+") && value.ne("v4.00+") => {
                return Err(ASSElementReaderError::Custom(format!(
                    "invalid value for key '{key:?}' in script info section: {value}"
                )))
            }
            ScriptInfoKey::WrapStyle if !("0".."3").contains(&value) => {
                return Err(ASSElementReaderError::Custom(format!(
                    "invalid value for key '{key:?}' in script info section: {value}"
                )))
            }
            key => key,
        };
        match self.script_info.get(&key) {
            Some(_) => Err(ASSElementReaderError::Custom(format!(
                "redefinition of key '{key:?}' in script info section"
            ))),
            None => {
                self.script_info.insert(key, value.to_string());
                Ok(())
            }
        }
    }

    fn read_v4_style(&mut self, line: &str) -> Result<(), ASSElementReaderError> {
        let line = if line.starts_with("Format:") {
            let line = line.split_once(':').unwrap().1.trim();
            self.styles_format = line
                .split(',')
                .flat_map(|str| str.trim().parse::<V4PlusStyleFields>())
                .collect();
            return Ok(());
        } else {
            line.split_once(':').unwrap().1.trim()
        };

        let fields: Vec<_> = line.trim().split(',').map(|str| str.trim()).collect();
        if fields.len() != self.styles_format.len() {
            return Err(ASSElementReaderError::Custom(format!(
                "style format specified {} fields, got {} in the style description",
                self.styles_format.len(),
                fields.len()
            )));
        }
        let fields: UnwrapHashMap<V4PlusStyleFields> =
            UnwrapHashMap(HashMap::from_iter(self.styles_format.iter().copied().zip(fields)));

        let encoding = fields.get(V4PlusStyleFields::Encoding);
        if encoding.ne("1") {
            return Err(ASSElementReaderError::Custom(format!(
                "we expected the encoding '1', got: {encoding}"
            )));
        }

        let name = fields.get(V4PlusStyleFields::Name);
        let style = ASSStyle {
            name: name.to_string(),
            font_name: fields.get(V4PlusStyleFields::Fontname).to_string(),
            font_size: fields.parsed(V4PlusStyleFields::Fontsize, parse_int, "font size")?,
            primary_color: fields.parsed(V4PlusStyleFields::PrimaryColour, parse_color, "primary")?,
            secondary_color: fields.parsed(V4PlusStyleFields::SecondaryColour, parse_color, "secondary")?,
            outline_color: fields.parsed(V4PlusStyleFields::OutlineColour, parse_color, "outline")?,
            back_color: fields.parsed(V4PlusStyleFields::BackColour, parse_color, "back")?,
            bold: fields.parsed(V4PlusStyleFields::Bold, parse_boolean, "bold")?,
            italic: fields.parsed(V4PlusStyleFields::Italic, parse_boolean, "italic")?,
            underline: fields.parsed(V4PlusStyleFields::Underline, parse_boolean, "underline")?,
            strikeout: fields.parsed(V4PlusStyleFields::StrikeOut, parse_boolean, "strikeout")?,
            scale_x: fields.parsed(V4PlusStyleFields::ScaleX, parse_float, "scale_x")?,
            scale_y: fields.parsed(V4PlusStyleFields::ScaleY, parse_float, "scale_y")?,
            spacing: fields.parsed(V4PlusStyleFields::Spacing, parse_float, "spacing")?,
            angle: fields.parsed(V4PlusStyleFields::Angle, parse_float, "angle")?,
            border_style: fields
                .get(V4PlusStyleFields::BorderStyle)
                .parse()
                .map_err(ASSElementReaderError::Custom)?,
            outline: fields.parsed(V4PlusStyleFields::Outline, parse_float, "outline")?,
            shadow: fields.parsed(V4PlusStyleFields::Shadow, parse_float, "shadow")?,
            alignment: fields
                .get(V4PlusStyleFields::Alignment)
                .parse()
                .map_err(ASSElementReaderError::Custom)?,
            margin_l: fields.parsed(V4PlusStyleFields::MarginL, parse_int, "margin_l")?,
            margin_r: fields.parsed(V4PlusStyleFields::MarginR, parse_int, "margin_r")?,
            margin_v: fields.parsed(V4PlusStyleFields::MarginV, parse_int, "margin_v")?,
        };

        match self.styles.insert(name.to_string(), style) {
            None => Ok(()),
            Some(old) => {
                log::error!(target: "ass", "redefine style '{name}', previous style was: {old:#?}");
                Err(ASSElementReaderError::Custom(format!(
                    "redefinition of style '{name}'"
                )))
            }
        }
    }

    fn read_event(&mut self, line: &str) -> Result<(), ASSElementReaderError> {
        let line = if line.starts_with("Format:") {
            let line = line.split_once(':').unwrap().1.trim();
            self.events_format = line
                .split(',')
                .flat_map(|str| str.trim().parse::<EventFields>())
                .collect();
            return match self.events_format.last() {
                Some(EventFields::Text) => Ok(()),
                _ => Err(ASSElementReaderError::Custom(format!(
                    "invalid format line: {:?}",
                    self.events_format
                ))),
            };
        } else {
            line.split_once(':').unwrap().1.trim()
        };
        let fields = line.splitn(self.events_format.len(), ',').collect::<Vec<_>>();
        if fields.len() != self.events_format.len() {
            return Err(ASSElementReaderError::Custom(format!(
                "style format specified {} fields, got {} in the style description",
                self.events_format.len(),
                fields.len()
            )));
        }
        let fields: UnwrapHashMap<EventFields> = UnwrapHashMap(HashMap::from_iter(
            self.events_format
                .iter()
                .copied()
                .zip(fields.into_iter().map(|s| s.trim())),
        ));
        self.events.push(ASSEvent {
            marked: fields.parsed_or(EventFields::Marked, parse_boolean, "marked", false),
            layer: fields.parsed_or(EventFields::Layer, parse_int, "layer", 0),
            start: fields.parsed(EventFields::Start, parse_date, "start")?,
            end: fields.parsed(EventFields::End, parse_date, "end")?,
            style: fields.get_or(EventFields::Style, "Default").to_string(),
            name: fields.get_or(EventFields::Name, "").to_string(),
            effect: fields.get_or(EventFields::Effect, "").to_string(),
            text: fields.get(EventFields::Text).to_string(),
        });
        Ok(())
    }
}

impl ASSElementReader for ASSReader {
    fn try_read(mut self, file: impl std::io::BufRead) -> Result<ASSContainer, ASSElementReaderError> {
        // Parse the file
        let mut skip_that_section = false;
        for line in file.lines() {
            let line = line.map_err(ASSElementReaderError::FailedToReadLine)?;
            let line = line.trim();

            if line.is_empty() || line.starts_with(';') {
                continue;
            } else if line.starts_with('[') && line.ends_with(']') {
                skip_that_section = false;
                self.section = match line.parse::<ASSFileSection>() {
                    Ok(section) => Some(section),
                    Err(err) => {
                        log::error!(target: "ass", "{err}");
                        skip_that_section = true;
                        continue;
                    }
                };
            } else if !skip_that_section {
                match self.section {
                    Some(ASSFileSection::ScriptInfo) => self.read_script_info(line)?,
                    Some(ASSFileSection::V4Styles) => self.read_v4_style(line)?,
                    Some(ASSFileSection::Events) => self.read_event(line)?,
                    None => {
                        return Err(ASSElementReaderError::Custom(format!(
                            "found the following line without a section: {line}"
                        )))
                    }
                }
            }
        }

        // Verify integrity + order events
        self.events.sort_by(|a, b| Ord::cmp(&a.start, &b.end));
        if !self.styles.contains_key(ASSStyle::default_name()) {
            self.styles.insert(ASSStyle::default_name_string(), ASSStyle::default());
        }
        for ASSEvent { style, .. } in &mut self.events {
            if !self.styles.contains_key(style) {
                *style = ASSStyle::default_name_string()
            }
        }

        // Convert events into lines...
        let mut lines = Vec::with_capacity(self.events.len());
        for event in self.events {
            let ASSEvent { start, end, .. } = event;
            lines.push(ASSLine { start, fini: end, ..Default::default() });
        }

        Ok(ASSContainer::from_parts(lines, self.script_info, self.styles))
    }
}
