use serde::{Deserialize, Serialize};
use vvs_procmacro::{EnumVariantCount, EnumVariantFromStr, EnumVariantIter};

/// Represents the types of the ASS types that can be manipulated. By combining them we can create
/// a tree of the ASS elements. We have:
/// ```ignore
/// - Lines
///   - content:
///     [0] := Line
///     ... ... ...
///     [n] := Line
/// - Line
///   - pos: AssPosition
///   - aux: HashMap<String, AssAuxValue>
///   - content: Syllabes
/// - Syllabes
///   - content:
///     [0] := Syllabe
///     ... ... ...
///     [n] := Syllabe
/// - Syllabe:
///   - pos: AssPosition
///   - aux: HashMap<String, AssAuxValue>
///   - content: String
/// ```
///
/// We also derive the Ord/PartialOrd crates. The types are ordered from the Lines to the Syllabe.
/// Thus if a type is greater than another one, the former must contains the latter.
#[derive(
    Debug,
    Clone,
    Copy,
    PartialEq,
    Eq,
    PartialOrd,
    Ord,
    Hash,
    Serialize,
    Deserialize,
    EnumVariantCount,
    EnumVariantIter,
    EnumVariantFromStr,
)]
pub enum ASSType {
    Lines = 0,
    Line = 1,
    Syllabes = 2,
    Syllabe = 3,
}

impl ASSType {
    /// Get the name of the ASS type.
    pub fn as_str(&self) -> &'static str {
        match self {
            ASSType::Lines => "lines",
            ASSType::Line => "line",
            ASSType::Syllabes => "syllabes",
            ASSType::Syllabe => "syllabe",
        }
    }

    /// Get the name of the type, but padded with spaces.
    pub fn as_padded_str(&self) -> &'static str {
        match self {
            ASSType::Lines => "lines   ",
            ASSType::Line => "line    ",
            ASSType::Syllabes => "syllabes",
            ASSType::Syllabe => "syllabe ",
        }
    }

    /// Returns the base type.
    pub fn base_type(&self) -> Self {
        match self {
            ASSType::Lines | ASSType::Line => ASSType::Line,
            ASSType::Syllabes | ASSType::Syllabe => ASSType::Syllabe,
        }
    }

    /// Returns the vec type.
    pub fn vec_type(&self) -> Self {
        match self {
            ASSType::Lines | ASSType::Line => ASSType::Lines,
            ASSType::Syllabes | ASSType::Syllabe => ASSType::Syllabes,
        }
    }
}

impl AsRef<str> for ASSType {
    fn as_ref(&self) -> &str {
        self.as_str()
    }
}

impl std::fmt::Display for ASSType {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_str(self.as_str())
    }
}
