use serde::{Deserialize, Serialize};
use std::{collections::HashMap, convert::TryFrom};

/// The values that can be added to an ASS element.
#[derive(Clone, PartialEq, Serialize, Deserialize)]
pub enum ASSAuxValue {
    Integer(i64),
    Floating(f64),
    Boolean(bool),
    String(String),
}

/// The auxiliary table of user values associated to ASS elements.
#[derive(Debug, Default, Clone, PartialEq)]
pub struct ASSAuxTable(HashMap<String, ASSAuxValue>);

impl std::fmt::Debug for ASSAuxValue {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Self::Integer(arg0) => write!(f, "{arg0}"),
            Self::Floating(arg0) => write!(f, "{arg0}"),
            Self::String(arg0) => write!(f, "{arg0:?}"),
            Self::Boolean(arg0) => write!(f, "{arg0}"),
        }
    }
}

impl ASSAuxValue {
    pub fn type_str(&self) -> &'static str {
        match self {
            ASSAuxValue::Floating(_) => "floating",
            ASSAuxValue::Integer(_) => "integer",
            ASSAuxValue::Boolean(_) => "boolean",
            ASSAuxValue::String(_) => "string",
        }
    }

    pub fn coerce_like(self, like: &ASSAuxValue) -> Option<ASSAuxValue> {
        use ASSAuxValue::*;
        match (&self, like) {
            (Floating(_), Floating(_))
            | (Integer(_), Integer(_))
            | (Boolean(_), Boolean(_))
            | (String(_), String(_)) => Some(self),

            (Integer(v), Floating(_)) => Some(Floating(i32::try_from(*v).ok()? as f64)),
            (Integer(0), Boolean(_)) => Some(Boolean(false)),
            (Integer(_), Boolean(_)) => Some(Boolean(true)),

            (Boolean(v), String(_)) => Some(String(format!("{v}"))),
            (Integer(v), String(_)) => Some(String(format!("{v}"))),
            (Floating(v), String(_)) => Some(String(format!("{v}"))),

            (Boolean(v), Integer(_)) => Some(Integer(*v as i64)),

            (String(_), Integer(_)) => todo!(),
            (String(_), Floating(_)) => todo!(),
            (String(_), Boolean(_)) => todo!(),

            (Floating(_), Integer(_)) | (Floating(_), Boolean(_)) | (Boolean(_), Floating(_)) => {
                log::error!(target: "lua", "invalid convertion from type `{}` to `{}`", self.type_str(), like.type_str());
                None
            }
        }
    }
}

impl From<String> for ASSAuxValue {
    fn from(value: String) -> Self {
        Self::String(value)
    }
}

impl From<&str> for ASSAuxValue {
    fn from(value: &str) -> Self {
        Self::String(value.to_string())
    }
}

impl From<bool> for ASSAuxValue {
    fn from(value: bool) -> Self {
        Self::Boolean(value)
    }
}

impl From<f32> for ASSAuxValue {
    fn from(value: f32) -> Self {
        Self::Floating(value as f64)
    }
}

impl From<f64> for ASSAuxValue {
    fn from(value: f64) -> Self {
        Self::Floating(value)
    }
}

impl From<i64> for ASSAuxValue {
    fn from(value: i64) -> Self {
        Self::Integer(value)
    }
}

impl From<i32> for ASSAuxValue {
    fn from(value: i32) -> Self {
        Self::Integer(value as i64)
    }
}

impl From<u32> for ASSAuxValue {
    fn from(value: u32) -> Self {
        Self::Integer(value as i64)
    }
}

impl From<i16> for ASSAuxValue {
    fn from(value: i16) -> Self {
        Self::Integer(value as i64)
    }
}

impl From<u16> for ASSAuxValue {
    fn from(value: u16) -> Self {
        Self::Integer(value as i64)
    }
}

impl From<i8> for ASSAuxValue {
    fn from(value: i8) -> Self {
        Self::Integer(value as i64)
    }
}

impl From<u8> for ASSAuxValue {
    fn from(value: u8) -> Self {
        Self::Integer(value as i64)
    }
}

impl std::fmt::Display for ASSAuxValue {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            ASSAuxValue::Floating(val) => write!(f, "{val}"),
            ASSAuxValue::Integer(val) => write!(f, "{val}"),
            ASSAuxValue::Boolean(val) => write!(f, "{val}"),
            ASSAuxValue::String(val) => f.write_str(val),
        }
    }
}

impl ASSAuxTable {
    pub fn new() -> Self {
        Default::default()
    }

    pub fn set(&mut self, name: impl AsRef<str>, value: ASSAuxValue) {
        let name = name.as_ref();
        let new = value.type_str();
        match self.0.get_mut(name) {
            Some(old) => match value.coerce_like(old) {
                Some(new) => *old = new,
                None => log::error!(
                    target: "lua",
                    "can't set new value for `{name}`, old value was of type `{}` and new one is of type `{new}`",
                    old.type_str()
                ),
            },
            None => {
                let _ = self.0.insert(name.to_string(), value);
            }
        }
    }

    pub fn get_copy(&self, name: impl AsRef<str>) -> Option<ASSAuxValue> {
        self.0.get(name.as_ref()).cloned()
    }

    pub fn get(&self, name: impl AsRef<str>) -> Option<&ASSAuxValue> {
        self.0.get(name.as_ref())
    }

    pub fn get_mut(&mut self, name: impl AsRef<str>) -> Option<&mut ASSAuxValue> {
        self.0.get_mut(name.as_ref())
    }
}

impl IntoIterator for ASSAuxTable {
    type Item = <HashMap<String, ASSAuxValue> as IntoIterator>::Item;
    type IntoIter = <HashMap<String, ASSAuxValue> as IntoIterator>::IntoIter;

    fn into_iter(self) -> Self::IntoIter {
        self.0.into_iter()
    }
}

impl FromIterator<(String, ASSAuxValue)> for ASSAuxTable {
    fn from_iter<T: IntoIterator<Item = (String, ASSAuxValue)>>(iter: T) -> Self {
        Self(HashMap::from_iter(iter))
    }
}
