use crate::parser::FileTypeValueParser;
use clap::Parser;
use clap_complete::Shell;
use std::path::PathBuf;
use vvs_utils::*;

fn get_cli_groups() -> impl IntoIterator<Item = clap::ArgGroup> {
    #[cfg(unix)]
    fn is_unix_target() -> bool {
        true
    }
    #[cfg(not(unix))]
    fn is_unix_target() -> bool {
        false
    }

    use clap::ArgGroup as grp;
    [
        grp::new("action").args(["manpage", "shell", "font-file", "script.vvs"]),
        grp::new("ass")
            .args(["subtitle.ass"])
            .conflicts_with_all(["manpage", "font-file"]),
        grp::new("opts")
            .args(["options.toml"])
            .conflicts_with_all(["shell", "manpage", "font-file"]),
        grp::new("infos")
            .args(["info"])
            .conflicts_with_all(["shell", "manpage", "font-file"]),
    ]
    .into_iter()
    .chain(either!(!is_unix_target() => None; Some(
        clap::ArgGroup::new("repl")  .args(["interactive"]) .conflicts_with_all(["shell", "manpage", "font-file"])
    )))
}

#[derive(Parser, Debug)]
#[command( author
         , version
         , about
         , name = "vvcc"
         , groups = get_cli_groups()
)]
pub struct Args {
    /// The script to run.
    ///
    /// A script must not touch to fields that begins by underscores, never. Those fields are
    /// reserved for vivy or for library developpers. Note that even library developpers must not
    /// touch or call fields from vivy that begins by underscores.
    #[arg( action       = clap::ArgAction::Set
         , id           = "script.vvs"
         , value_parser = FileTypeValueParser::new("vvs")
    )]
    pub script: Option<PathBuf>,

    /// The input ASS file to execute the script on.
    ///
    /// The supported subtitle extensions are the ASS files (V4+) or the Json files exported by the
    /// Vivy application.
    #[arg ( short        = 'f'
          , long         = "subtitle"
          , action       = clap::ArgAction::Set
          , id           = "subtitle.ass"
          , value_parser = FileTypeValueParser::new("ass"),
    )]
    pub ass_file: Option<PathBuf>,

    /// The option file that will be used when running the script.
    #[arg( short        = 't'
         , long         = "option"
         , action       = clap::ArgAction::Set
         , id           = "options.toml"
         , value_parser = FileTypeValueParser::new("toml")
    )]
    pub options: Option<PathBuf>,

    /// The includes' folder list, in order.
    ///
    /// Will search for modules to import in those folders. Note that there is no need to add the
    /// folder of the loaded script as it will automatically be added in the top position of the
    /// include list.
    #[arg( short  = 'I'
         , long   = "include-path"
         , action = clap::ArgAction::Append
         , id     = "include"
    )]
    pub include_folders: Vec<PathBuf>,

    /// Launch vvcc REPL after loading the passed script if any.
    ///
    /// In REPL mode you must not touch to fields that begins by underscores, never. Those fields
    /// are reserved for vivy or for library developpers. Note that even library developpers must
    /// not touch or call fields from vivy that begins by underscores.
    #[cfg(unix)]
    #[arg( short  = 'i'
         , action = clap::ArgAction::SetTrue
    )]
    pub interactive: bool,

    /// Shows informations about a script.
    ///
    /// The informations consists of the loaded modules, all the options, their possible values,
    /// the order of the jobs, etc.
    ///
    /// When printing informations about a script you won't be able to run it. You can however
    /// print the informations about a script then enter in iteractive mode.
    #[arg( short  = 'P'
         , long   = "info"
         , action = clap::ArgAction::SetTrue
    )]
    pub info: bool,

    /// Display infos about embeded fonts or external fonts.
    #[arg( long   = "font-info"
         , action = clap::ArgAction::Set
         , id     = "font-file"
    )]
    pub font_info: Option<Option<PathBuf>>,

    /// Make vvcc more verbose, repeat to make it even more verbose
    #[arg( long
         , short  = 'v'
         , action = clap::ArgAction::Count
    )]
    pub verbose: u8,

    /// Generate the manpage for vvcc and print it out to stdout
    ///
    /// You may read it by running `vvcc --manpage | man -l -` or writing it in the
    /// `$HOME/.local/share/man/man1/` or `/usr/share/man/man1` folder
    #[arg( long   = "manpage"
         , action = clap::ArgAction::SetTrue
    )]
    pub manpage: bool,

    /// Generate completion script for the given shell
    #[arg( long   = "shell"
         , action = clap::ArgAction::Set
    )]
    pub shell: Option<Shell>,
}
