use crate::args::Args;
use clap_complete::Shell;
use serde::{Deserialize, Serialize};
use std::path::PathBuf;
use vvs_utils::*;

#[derive(Debug, Serialize, Deserialize)]
pub struct ConfigKaraMaker {
    #[serde(rename = "name")]
    kara_maker: String,

    email: Option<String>,
}

#[derive(Debug, Default, Serialize, Deserialize)]
pub struct ConfigFile {
    #[serde(rename = "includes")]
    include_folders: Vec<PathBuf>,

    karamaker: ConfigKaraMaker,
}

#[derive(Debug, Default)]
pub struct Config {
    // From config::ConfigFile
    pub kara_maker: String,

    // From args::Args
    pub script: Option<PathBuf>,
    pub ass_file: Option<PathBuf>,
    pub options: Option<PathBuf>,
    pub interactive: bool,
    pub info: bool,
    pub manpage: bool,
    pub shell: Option<Shell>,
    pub font_info: Option<Option<PathBuf>>,
    pub verbose: u8,

    // Merge of the two
    pub include_folders: Vec<PathBuf>,
}

macro_rules! merge {
    ($($what: ident $src: expr => $dest: expr, { $( $field: ident ),+ });+ $(;)?) => {{
        $(merge! { @@internal $what $( $field ),+ : $src => $dest });+;
    }};

    (@@internal override $( $field: ident ),+ : $src: expr => $dest: expr) => {{$( $dest.$field  = $src.$field; )+}};
    (@@internal flg_or   $( $field: ident ),+ : $src: expr => $dest: expr) => {{$( $dest.$field |= $src.$field; )+}};
    (@@internal flg_and  $( $field: ident ),+ : $src: expr => $dest: expr) => {{$( $dest.$field &= $src.$field; )+}};

    (@@internal append   $( $field: ident ),+ : $src: expr => $dest: expr) => {{$( $dest.$field.extend($src.$field); )+}};

    (@@internal max      $( $field: ident ),+ : $src: expr => $dest: expr) => {{$( $dest.$field = max_partial!($src.$field, $dest.$field); )+}};
    (@@internal min      $( $field: ident ),+ : $src: expr => $dest: expr) => {{$( $dest.$field = min_partial!($src.$field, $dest.$field); )+}};

    (@@internal set_if_not $( $field: ident),+ : $src: expr => $dest: expr) => {{$(
        if $dest.$field.is_none() {
            $dest.$field = $src.$field;
        }
    )+}};
}

impl Extend<ConfigFile> for ConfigFile {
    fn extend<T: IntoIterator<Item = ConfigFile>>(&mut self, iter: T) {
        iter.into_iter().for_each(|other| {
            let (km, self_km) = (other.karamaker, &mut self.karamaker);
            merge! {
                append   other => self,    { include_folders };
                override km    => self_km, { kara_maker };
            }
        });
    }
}

impl Extend<ConfigFile> for Config {
    fn extend<T: IntoIterator<Item = ConfigFile>>(&mut self, iter: T) {
        iter.into_iter().for_each(|cfg| {
            let km = cfg.karamaker;
            merge! {
                append   cfg => self, { include_folders };
                override km  => self, { kara_maker };
            }
        });
    }
}

impl Extend<Args> for Config {
    fn extend<T: IntoIterator<Item = Args>>(&mut self, iter: T) {
        iter.into_iter().for_each(|args| {
            merge! {
                append     args => self, { include_folders };
                set_if_not args => self, { script, ass_file, options, shell, font_info };
                max        args => self, { verbose };
                override   args => self, { info, manpage, interactive };
            }
        });
    }
}

impl ConfigFile {
    pub fn serialize(&self) -> Result<String, Box<dyn std::error::Error>> {
        Ok(toml::to_string_pretty(self).map_err(Box::new)?)
    }

    pub fn deserialize(input: String) -> Result<Self, Box<dyn std::error::Error>> {
        Ok(toml::from_str(&input).map_err(Box::new)?)
    }
}

impl Default for ConfigKaraMaker {
    fn default() -> Self {
        Self { kara_maker: "Viieux".to_string(), email: Default::default() }
    }
}
