use log::{Level, Metadata, Record, SetLoggerError};
use std::sync::{atomic::AtomicU8, OnceLock};
use thiserror::Error;

#[derive(Debug, Error)]
pub struct LoggerInitError(SetLoggerError);

#[derive(Debug, Default)]
struct SimpleLogger {
    level: AtomicU8,
}

#[repr(transparent)]
#[derive(Debug, Default)]
struct SimpleLoggerRef {
    inner: SimpleLogger,
}

static LOGGER: OnceLock<SimpleLoggerRef> = OnceLock::new();

fn logger() -> &'static SimpleLogger {
    &LOGGER.get().unwrap().inner
}

impl std::fmt::Display for LoggerInitError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.0)
    }
}

impl SimpleLogger {
    fn write_str<S: AsRef<str>>(level: char, prefix: String, target: &str, content: S) {
        let prefix = if target.is_empty() {
            format!("{level} [{prefix}]")
        } else if prefix.is_empty() {
            format!("{level} [{target}] ")
        } else {
            format!("{level} [{prefix} {target}] ")
        };
        let mut content = content.as_ref().lines().filter(|content| !content.trim().is_empty());
        if let Some(content) = content.next() {
            eprintln!("{prefix}{content}");
        }
        content.for_each(|content| eprintln!("{level} {content}"));
    }

    fn level(&self) -> Level {
        // Always display errors and warnings.
        match self.level.load(std::sync::atomic::Ordering::SeqCst) {
            0 => Level::Warn,
            1 => Level::Info,
            2 => Level::Debug,
            _ => Level::Trace,
        }
    }
}

impl log::Log for SimpleLogger {
    fn enabled(&self, metadata: &Metadata) -> bool {
        metadata.level() <= self.level()
    }

    fn log(&self, record: &Record) {
        if self.enabled(record.metadata()) {
            let level = match record.level() {
                Level::Error => '!',
                Level::Warn => '*',
                Level::Info => '#',
                Level::Debug => '.',
                Level::Trace => ' ',
            };
            let prefix = match (record.file(), record.line()) {
                (None, None) => "".to_string(),
                (None, Some(line)) => format!("...+{line}"),
                (Some(file), None) => format!("{file}+..."),
                (Some(file), Some(line)) => format!("{file}+{line}"),
            };
            if let Some(s) = record.args().as_str() {
                SimpleLogger::write_str(level, prefix, record.target(), s)
            } else {
                SimpleLogger::write_str(level, prefix, record.target(), record.args().to_string());
            }
        }
    }

    fn flush(&self) {}
}

pub fn level(lvl: u8) {
    logger().level.store(lvl, std::sync::atomic::Ordering::SeqCst);
    log::set_max_level(LOGGER.get().unwrap().inner.level().to_level_filter());
}

pub fn init(level: Option<Level>) -> Result<(), LoggerInitError> {
    LOGGER.set(Default::default()).expect("failed to set default logger...");
    log::set_logger(logger())
        .map(|()| {
            log::set_max_level(match level {
                None => logger().level().to_level_filter(),
                Some(level) => {
                    match level {
                        Level::Trace => self::level(3),
                        Level::Debug => self::level(2),
                        Level::Info => self::level(1),
                        Level::Warn | Level::Error => self::level(0),
                    };
                    level.to_level_filter()
                }
            });
        })
        .map_err(LoggerInitError)
}
