//! The VivyScript cli
#![forbid(unsafe_code)]

use anyhow::{Context, Result};
use vvs_cli::{
    args,
    config::{Config, ConfigFile},
    logger,
};
use vvs_utils::xdg::*;

fn print_face_info(name: &str, font: &[u8]) -> Result<()> {
    let font = vvs_font::Font::try_from(font).with_context(|| format!("failed to parse font {name}"))?;

    let font_types = font
        .font_types()
        .into_iter()
        .map(|ty| format!("{ty:?}"))
        .collect::<Vec<_>>()
        .join(", ");

    println!("###");
    println!(
        "# Family Name:        {}",
        font.name().with_context(|| "failed to get the font name")?
    );
    println!("# Name(s):            {}", font.family_names().join(", "));
    println!("# Font Type(s):       {font_types}");
    println!("# Number of glyph(s): {}", font.number_of_glyphs());

    Ok(())
}

fn main() -> Result<()> {
    logger::init(None).map_err(Box::new)?;

    let mut config = Config::default();
    config.extend([
        XDGConfig::<ConfigFile, XDGConfigMergedSilent>::new("vvcc", ConfigFile::deserialize)
            .file("vvcc.toml")
            .read_or_default(ConfigFile::serialize),
    ]);
    config.extend([<args::Args as clap::Parser>::parse()]);
    let Config {
        script,
        ass_file: _,
        options: _,
        #[cfg(unix)]
        interactive,
        info: _,
        manpage,
        shell,
        font_info,
        verbose,
        include_folders: _,
        ..
    } = config;
    #[cfg(not(unix))]
    let interactive = false;
    logger::level(verbose);

    if manpage {
        log::debug!(target: "vvcc", "generate the manpage for vvcc");
        use clap::CommandFactory;
        return clap_mangen::Man::new(args::Args::command())
            .render(&mut std::io::stdout())
            .with_context(|| "failed to render manpage for vvcc");
    } else if let Some(shell) = shell {
        log::debug!(target: "vvcc", "generate shell completion script for shell {shell}");
        use clap::CommandFactory;
        let mut command = args::Args::command();
        let command_name = command.get_name().to_string();
        clap_complete::generate(shell, &mut command, command_name, &mut std::io::stdout());
        return Ok(());
    } else if let Some(font_info) = font_info {
        match font_info {
            Some(path) => {
                let font =
                    std::fs::read(&path).with_context(|| format!("failed to read font: {}", path.to_string_lossy()))?;
                print_face_info(&path.to_string_lossy(), &font)?;
            }
            None => {
                for (name, font) in vvs_font::embeded_fonts() {
                    print_face_info(name, font)?;
                }
            }
        }
        return Ok(());
    }

    if script.is_none() && !interactive {
        use clap::CommandFactory;
        return args::Args::command()
            .print_help()
            .with_context(|| "failed to print help message for vvcc");
    }

    Ok(())
}
