use crate::{error::*, *};
use ab_glyph::Font as _;

/// The different types of fonts. To be sure of what you are doing, please use only regular
/// fonts, behaviour of non regular fonts may change depending on the platform if the italic or
/// bold tag is set in the ASS Style.
#[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord)]
pub enum FontType {
    Regular,
    Italic,
    Bold,
    Oblique,
    Monospaced,
    Variable,
}

/// Struct used to store informations we need about a font.
#[derive(Debug)]
pub struct Font<'a> {
    font: ab_glyph::FontRef<'a>,
    face: ttf_parser::Face<'a>,
}

impl<'a> TryFrom<&'a [u8]> for Font<'a> {
    type Error = FontCreationError;

    fn try_from(data: &'a [u8]) -> Result<Self, Self::Error> {
        Ok(Self {
            face: ttf_parser::Face::parse(data, 0).map_err(FontCreationError::TTFParserError)?,
            font: ab_glyph::FontRef::try_from_slice(data).map_err(FontCreationError::ABGlyphError)?,
        })
    }
}

impl<'a> Font<'a> {
    /// Get the general name of the font.
    pub fn name(&self) -> Result<String, FontError> {
        self.face
            .names()
            .into_iter()
            .find(|name| name.name_id == ttf_parser::name_id::FAMILY && name.is_unicode())
            .and_then(|name| name.to_string())
            .ok_or(FontError::NoName)
    }

    /// Get the family names.
    pub fn family_names(&self) -> Vec<String> {
        let names = self.face.names();
        let filter = [ttf_parser::name_id::POST_SCRIPT_NAME, ttf_parser::name_id::FULL_NAME];
        names
            .into_iter()
            .flat_map(|name| (filter.contains(&name.name_id) && name.is_unicode()).then(|| name.to_string()))
            .flatten()
            .collect()
    }

    /// The the font types.
    pub fn font_types(&self) -> Vec<FontType> {
        use FontType::*;
        let mut ret = [
            self.face.is_regular().then_some(Regular),
            self.face.is_italic().then_some(Italic),
            self.face.is_bold().then_some(Bold),
            self.face.is_oblique().then_some(Oblique),
            self.face.is_monospaced().then_some(Monospaced),
            self.face.is_variable().then_some(Variable),
        ]
        .into_iter()
        .flatten()
        .chain(Some(match self.face.style() {
            ttf_parser::Style::Normal => Regular,
            ttf_parser::Style::Italic => Italic,
            ttf_parser::Style::Oblique => Oblique,
        }))
        .collect::<Vec<_>>();
        ret.sort();
        ret.dedup();
        ret
    }

    /// Get the number of glyphs in the font.
    pub fn number_of_glyphs(&self) -> i64 {
        self.face.number_of_glyphs() as i64
    }

    /// Outline a glyph with the specified pt size.
    pub fn outline_glyph(&self, pt: f64, glyph: char) -> Result<Rect, FontError> {
        self.font
            .outline_glyph(
                self.font.glyph_id(glyph).with_scale(
                    self.font
                        .pt_to_px_scale(f64::clamp(pt, f32::MIN.into(), f32::MAX.into()) as f32)
                        .expect("failed to get the px_scale..."),
                ),
            )
            .map(|outlined| {
                let ab_glyph::Rect { min, max } = outlined.px_bounds();
                Rect::new(
                    Point { x: min.x.trunc() as i64, y: min.y.trunc() as i64 },
                    Point { x: max.x.trunc() as i64, y: max.y.trunc() as i64 },
                )
            })
            .ok_or(FontError::FailedToOutline(pt, glyph))
    }

    /// Outline a string slice with the specified pt size.
    pub fn outline_str(&self, pt: f64, str: impl AsRef<str>) -> Result<Rect, FontError> {
        let (rects, errs): (Vec<_>, Vec<_>) = str
            .as_ref()
            .chars()
            .map(|glyph| self.outline_glyph(pt, glyph))
            .partition(Result::is_ok);
        if let Some(err) = errs.into_iter().next() {
            return err;
        }
        let mut rects = rects.into_iter().map(Result::unwrap);
        let first = rects.next().ok_or(FontError::EmptyStringToOutline)?;
        Ok(rects.fold(first, Rect::merge))
    }
}
