use std::cmp::{max, min};

/// Describes coordinates in the plan, the origin is the top left corner of the screen.
#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub struct Point {
    pub x: i64,
    pub y: i64,
}

/// Describes an area in the plan, the origin is the top left corner of the screen.
#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub struct Rect {
    top_left_corner: Point,
    bottom_right_corner: Point,
}

impl Rect {
    /// Create a correctly formed [Rect] that includes the passed [Point].
    pub fn new(p1: Point, p2: Point) -> Self {
        Self {
            top_left_corner: Point::min(p1, p2),
            bottom_right_corner: Point::max(p1, p2),
        }
    }

    /// Returns a new [Rect] that includes the passed rectangles.
    pub fn merge(self, other: Self) -> Self {
        Self {
            top_left_corner: Point::min(self.top_left_corner, other.top_left_corner),
            bottom_right_corner: Point::min(self.bottom_right_corner, other.bottom_right_corner),
        }
    }

    /// Get the Top Left corner, the min coordinates.
    pub fn tl_corner(&self) -> Point {
        self.top_left_corner
    }

    /// Get the Bottom Right corner, the max coordinates.
    pub fn br_corner(&self) -> Point {
        self.bottom_right_corner
    }
}

impl Point {
    /// Returns the min components of the two [Point].
    pub fn min(p1: Point, p2: Point) -> Point {
        Point { x: min(p1.x, p2.x), y: min(p1.y, p2.y) }
    }

    /// Returns the max components of the two [Point].
    pub fn max(p1: Point, p2: Point) -> Point {
        Point { x: max(p1.x, p2.x), y: max(p1.y, p2.y) }
    }
}
