/*
BSD License

Copyright (c) 2013, Kazunori Sakamoto
Copyright (c) 2016, Alexander Alexeev
Copyright (c) 2023, Maël Martin
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
3. Neither the NAME of Rainer Schuster nor the NAMEs of its contributors may be
   used to endorse or promote products derived from this software without
   specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

grammar VVL;

chunk:      topblock* EOF;
visibility: ('pub')?;
block:      stat* laststat?;

topblock: visibility 'function' NAME (':' NAME)? funcbody
        | visibility 'job'      NAME funcbody
        | visibility 'const'    NAME opttypespec '=' exp
        | visibility 'option'   NAME opttypespec '=' exp
        | ('import' | 'requires') string
        ;

stat: ';'
    | varlist '=' explist
    | functioncall
    | 'do' block 'end'
    | 'while' exp 'do' block 'end'
    | 'repeat' block 'until' exp
    | 'if' exp 'then' block ('elseif' exp 'then' block)* ('else' block)? 'end'
    | 'for' NAME '=' exp ',' exp (',' exp)? 'do' block 'end'
    | 'for' namelist 'in' explist 'do' block 'end'
    | 'let' attnamelist ('=' explist)?
    ;

attnamelist:    NAME opttypespec (',' NAME opttypespec)*;
opttypespec:    (':' NAME)?;
laststat:       'return' explist? | 'break' | 'continue' ';'?;
varlist:        var (',' var)*;
namelist:       NAME (',' NAME)*;
explist:        (exp ',')* exp;

exp: 'nil' | 'false' | 'true'
   | number
   | color
   | string
   | prefixexp
   | tableconstructor
   | <assoc=right> exp operatorPower exp
   | operatorUnary exp
   | exp operatorMulDivMod exp
   | exp operatorAddSub exp
   | <assoc=right> exp operatorStrcat exp
   | exp operatorComparison exp
   | exp operatorAnd exp
   | exp operatorOr exp
   | exp operatorBitwise exp
   | '(' exp (',' exp)+ (',')? ')'
   ;

prefixexp:          varOrExp nameAndArgs*;
functioncall:       varOrExp nameAndArgs+;
varOrExp:           var | '(' exp ')';
var:                (NAME | '(' exp ')' varSuffix) varSuffix*;
varSuffix:          nameAndArgs* ('[' exp ']' | '.' NAME);
nameAndArgs:        (':' NAME)? args;
args:               '(' explist? ')' | tableconstructor | string;
funcbody:           '(' parlist? ')' '->' NAME 'begin' block 'end';
parlist:            namelist (',')?;
tableconstructor:   '{' fieldlist? '}';
fieldlist:          field (fieldsep field)* fieldsep?;
field:              '[' exp ']' '=' exp | NAME '=' exp | exp;
fieldsep:           ',' | ';';

operatorOr:         'or';
operatorAnd:        'and';
operatorComparison: '<' | '>' | '<=' | '>=' | '~=' | '==' | '!=';
operatorStrcat:     '..';
operatorAddSub:     '+' | '-';
operatorMulDivMod:  '*' | '/' | '%' | '//' | 'mod';
operatorBitwise:    '&' | '^' | '|' | '~' | '<<' | '>>';
operatorUnary:      'not' | '#' | '-' | '~';
operatorPower:      '^';

color:  '#(' ... ')';
number: INT | HEX | FLOAT | HEX_FLOAT;
string: NORMALSTRING | CHARSTRING | LONGSTRING;

// LEXER

NAME:           [a-zA-Z_][a-zA-Z_0-9]*;
NORMALSTRING:   '"' ( EscapeSequence | ~('\\'|'"') )* '"';
CHARSTRING:     '\'' ( EscapeSequence | ~('\''|'\\') )* '\'';
LONGSTRING:     '[' NESTED_STR ']';

fragment
NESTED_STR: '=' NESTED_STR '='
          | '[' .*? ']'
          ;

INT: Digit+;
HEX: '0' [xX] HexDigit+;

FLOAT: Digit+ '.' Digit* ExponentPart?
     | '.' Digit+ ExponentPart?
     | Digit+ ExponentPart
     ;

HEX_FLOAT
    : '0' [xX] HexDigit+ '.' HexDigit* HexExponentPart?
    | '0' [xX] '.' HexDigit+ HexExponentPart?
    | '0' [xX] HexDigit+ HexExponentPart
    ;

fragment
ExponentPart
    : [eE] [+-]? Digit+
    ;

fragment
HexExponentPart
    : [pP] [+-]? Digit+
    ;

fragment
EscapeSequence
    : '\\' '\r'? '\n'
    ;

fragment
Digit
    : [0-9]
    ;

fragment
HexDigit
    : [0-9a-fA-F]
    ;

fragment
SingleLineInputCharacter
    : ~[\r\n\u0085\u2028\u2029]
    ;

COMMENT
    : '--[' NESTED_STR ']' -> channel(HIDDEN)
    ;

LINE_COMMENT
    : '--' SingleLineInputCharacter* -> channel(HIDDEN)
    ;

WS
    : [ \t\u000C\r\n]+ -> skip
    ;
