use crate::{ast::*, parser::*};
use hashbrown::{HashMap, HashSet};
use std::{cell::RefCell, rc::Rc};

/// The first element of the tuple is the destination variable, the second is the tuple to describe
/// the job picked from a module, the last is the list of variables to use as input for this job.
pub type ProgramOperation = (ASTString, (ASTString, ASTString), Vec<ASTString>);

/// The main VVS file/program.
pub struct Program {
    modules: HashMap<ASTString, Module>,

    /// Options specified in the VVL module. All options must be constants.
    ///
    /// TODO: Remove that, use the options from the modules directly...
    options: Vec<((ASTString, ASTVar), ASTConst)>,

    /// The setted options, from the VVS file.
    setted: Vec<((ASTString, ASTString), ASTConst)>,

    /// The operations to do in order.
    operations: Vec<ProgramOperation>,

    /// What the program writes.
    writes: (ASTSpan, Vec<ASTString>),

    /// Initial variable.
    initial_var: ASTString,

    /// Caching for strings, identified by theyr hash and reuse the same memory location to reduce
    /// memory footprint. Use the .finish function from the hasher to get the key.
    strings: Rc<RefCell<HashMap<u64, ASTString>>>,
}

/// Here we have all the getters...
impl Program {
    /// Get a handle to the string cache.
    pub fn strings(&mut self) -> ASTStringCacheHandle {
        ASTStringCacheHandle::new(self.strings.borrow_mut())
    }

    /// Get a const reference to a module.
    pub fn module(&self, name: impl AsRef<str>) -> Option<&Module> {
        self.modules.get(name.as_ref())
    }

    /// Get a mutable reference to a module.
    pub fn module_mut(&mut self, name: impl AsRef<str>) -> Option<&mut Module> {
        self.modules.get_mut(name.as_ref())
    }

    /// Get the value for an option. For the resolution: we first take the value from the setted
    /// list before the default value. Note that if we set an option that was never declared we
    /// don't return it here, it should have raised a warning and should not be used in the code
    /// anyway.
    pub fn options(&self, module: &ASTString, name: &ASTString) -> Option<&ASTConst> {
        let default = self
            .options
            .iter()
            .find_map(|((mm, var), val)| (*module == *mm && *name == *var.name()).then_some(val))?;
        let setted = self
            .setted
            .iter()
            .find_map(|((mm, var), val)| (*module == *mm && *name == *var).then_some(val));
        Some(setted.unwrap_or(default))
    }
}

/// Here we have all the setters... We always take the [Program] and return a new one for the
/// functions to be pure.
impl Program {
    /// Get the operations and the variables to write. Can return an error if a variable was
    /// assigned multiple times or if a variable to write was never assigned.
    pub fn into_operations(self) -> VVResult<(ASTString, Vec<ProgramOperation>, Vec<ASTString>)> {
        let assigned = HashSet::<ASTString>::from_iter(
            self.operations
                .iter()
                .map(|(dest, ..)| dest.clone())
                .chain([self.initial_var.clone()]),
        );
        if let Some(unwritten) = self.writes.1.iter().find(|var| !assigned.contains(*var)) {
            return Err(VVError::ErrorMsg(
                self.writes.0,
                format!("variable `{unwritten}` can't be saved because it was never assigned"),
            ));
        }
        let res = (self.initial_var, self.writes);
        todo!("add the verified operations to {res:?}")
    }

    /// Set an option in the [Program], if the option was already set or was not declared in a
    /// [Module] we raise a warning.
    pub fn set_option(mut self, module: ASTString, name: ASTString, value: ASTConst) -> Self {
        if self
            .options
            .iter()
            .any(|((mm, nn), _)| *mm == module && name == *nn.name())
        {
            log::warn!(target: "cc", ";set option {module}.{name} which was not declared");
        }
        if self.setted.iter().any(|((mm, nn), _)| *mm == module && name == *nn) {
            log::warn!(target: "cc", ";re-set option {module}.{name}");
        }
        self.setted.push(((module, name), value));
        self
    }

    /// Add an option to the module. We try to expose a functional semantic for the usage of the
    /// [Program] struct to help the parsing implementation.
    pub fn declare_option(mut self, module: ASTString, var: ASTVar, value: ASTConst) -> VVResult<Self> {
        let key = (module, var);
        if let Some(old) = self.options.iter().find_map(|(k, _)| key.eq(k).then_some(k.1 .0)) {
            Err(VVError::OptionRedefined(old, key.0, key.1))
        } else {
            self.options.push((key, value));
            Ok(self)
        }
    }
}

impl std::fmt::Debug for Program {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("Program")
            .field("modules", &self.modules)
            .field("options", &self.options)
            .field("setted", &self.setted)
            .field("operations", &self.operations)
            .field("writes", &self.writes)
            .field("initial_var", &self.initial_var)
            .finish()
    }
}
