/// A span, without the borrowed string.
#[derive(Debug, Default, Clone, Copy, PartialEq, Eq, Hash)]
pub struct ASTSpan {
    line: u64,
    column: u64,
    offset: u64,
}

impl std::fmt::Display for ASTSpan {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let Self { line, column, .. } = &self;
        write!(f, "+{line}:{column}")
    }
}

impl ASTSpan {
    /// Create the span out of a line and column and an offset. The lines and
    /// columns begeins at 1. The offset begins at 0.
    pub fn new(line: u64, column: u64, offset: u64) -> Self {
        assert!(line >= 1);
        assert!(column >= 1);
        Self { line, column, offset }
    }

    /// Get the column of the span. The column starts at 1
    pub fn line(&self) -> u64 {
        self.line
    }

    /// Get the column of the span. The column starts at 1
    pub fn column(&self) -> u64 {
        self.column
    }

    /// Get the offset of the span.
    pub fn offset(&self) -> u64 {
        self.offset
    }

    /// Merge two spans. For now we just take the first one (the one with the
    /// lesser offset, i.e. the minimal one). In the future we will update the
    /// length field (when it's added to the structure...)
    pub fn merge(s1: Self, s2: Self) -> Self {
        if PartialOrd::gt(&s1, &s2) {
            s2
        } else {
            s1
        }
    }
}

impl From<nom_locate::LocatedSpan<&str>> for ASTSpan {
    fn from(span: nom_locate::LocatedSpan<&str>) -> Self {
        Self {
            line: span.location_line() as u64,
            column: span.naive_get_utf8_column() as u64,
            offset: span.location_offset() as u64,
        }
    }
}

impl PartialOrd for ASTSpan {
    fn partial_cmp(&self, other: &Self) -> Option<std::cmp::Ordering> {
        PartialOrd::partial_cmp(&self.offset, &other.offset)
    }
}
