use super::ASTString;
use hashbrown::HashMap;
use std::{cell::RefMut, collections::hash_map::DefaultHasher, hash::Hasher};

/// Used when iterating into the module with mut access, we might need to access the string
/// cache...
#[derive(Debug)]
pub struct ASTStringCacheHandle<'a> {
    strings: RefMut<'a, HashMap<u64, ASTString>>,
}

impl<'a> ASTStringCacheHandle<'a> {
    pub(crate) fn new(strings: RefMut<'a, HashMap<u64, ASTString>>) -> Self {
        Self { strings }
    }

    /// Get the id of a string.
    fn get_id(str: impl AsRef<str>) -> u64 {
        let mut hasher = DefaultHasher::new();
        hasher.write(str.as_ref().as_bytes());
        hasher.finish()
    }

    /// Get a string from the cache, fail if not present.
    pub fn get(&mut self, str: impl AsRef<str>) -> Option<ASTString> {
        self.strings.get(&Self::get_id(str.as_ref())).cloned()
    }

    /// Get or create a string in the cache.
    pub fn get_or_insert(&mut self, str: impl AsRef<str>) -> ASTString {
        let id = Self::get_id(str.as_ref());
        match self.strings.get(&id) {
            Some(str) => str.clone(),
            None => {
                let str = ASTString::from(str.as_ref());
                self.strings.insert(id, str.clone());
                str
            }
        }
    }
}
