use super::ASTSpan;
use hashbrown::HashMap;
use std::rc::Rc;

pub type ASTString = Rc<str>;
pub type ASTFloating = f32;
pub type ASTInteger = i32;
pub type ASTTable<T> = HashMap<ASTString, T>;

#[macro_export]
macro_rules! anon_expression {
    ($variant: ident $($args: tt)?) => {
        ASTExpr {
            span: Default::default(),
            content: ASTExprVariant::$variant $($args)?,
        }
    };
}

#[macro_export]
macro_rules! expression {
    ($span: expr, $variant: ident $($args: tt)?) => {
        ASTExpr {
            span: $span.into(),
            content: ASTExprVariant::$variant $($args)?,
        }
    };
}

#[macro_export]
macro_rules! anon_instruction {
    ($variant: ident $($args: tt)?) => {
        ASTInstr {
            span: Default::default(),
            content: ASTInstrVariant::$variant $($args)?,
        }
    };
}

#[macro_export]
macro_rules! instruction {
    ($span: expr, $variant: ident $($args: tt)?) => {
        ASTInstr {
            span: $span.into(),
            content: ASTInstrVariant::$variant $($args)?,
        }
    };
}

/// Job to execute on a line, syllabe, list of lines, etc.
#[derive(Debug)]
pub struct ASTJob {
    pub name: ASTString,
    pub returns: ASTType,
    pub arguments: Vec<ASTVar>,
    pub content: Vec<ASTInstr>,

    pub span: ASTSpan,
}

/// A function.
#[derive(Debug)]
pub struct ASTFunction {
    pub name: ASTString,
    pub arguments: Vec<ASTVar>,
    pub content: Vec<ASTInstr>,

    pub span: ASTSpan,
}

/// Instructions.
#[derive(Debug)]
pub struct ASTInstr {
    pub content: ASTInstrVariant,
    pub span: ASTSpan,
}

/// Instructions.
#[derive(Debug, PartialEq)]
pub enum ASTInstrVariant {
    /// Declare variables.
    Decl(Vec<ASTVar>, Vec<ASTExpr>),

    /// Assign into variables.
    Assign(Vec<ASTExpr>, Vec<ASTExpr>),

    /// FunctionCall.
    FuncCall(ASTString, Vec<ASTExpr>),

    /// Begin a block.
    Block(Vec<ASTInstr>),

    /// A WhileDo instruction.
    WhileDo(ASTExpr, Vec<ASTInstr>),

    /// A DoWhile instruction.
    RepeatUntil(ASTExpr, Vec<ASTInstr>),

    /// Conditionals, contract the elseid blocks.
    Cond {
        cond: ASTExpr,
        then_block: Vec<ASTInstr>,
        elseif_blocks: Vec<(ASTExpr, Vec<ASTInstr>)>,
        else_block: Option<Vec<ASTInstr>>,
    },

    /// For loop, the classic one:
    /// ```vvs
    /// for elem = 1, 3, 1 do print(elem) end
    /// ```
    ForLoop {
        var: ASTVar,
        lower: ASTInteger,
        upper: ASTInteger,
        step: Option<ASTInteger>,
    },

    /// For loop with an iterable expression (table):
    /// ```vvs
    /// for elem in { 1, 2, 3 } do print(elem) end
    /// for elem in ( 1, 2, 3 ) do print(elem) end
    /// for elem in my_table    do print(elem) end
    /// ```
    ForInto { var: ASTVar, list: ASTExpr },

    /// Final thing: break from block.
    Break,

    /// Final thing: continue to next iteration.
    Continue,

    /// Final thing: return something.
    Return(ASTExpr),
}

/// Binops, sorted by precedence.
#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub enum ASTBinop {
    /// Assoc: right
    Power,

    Mul,
    Div,
    Mod,
    Add,
    Sub,

    /// Assoc: right
    StrCat,

    CmpLE,
    CmpLT,
    CmpGE,
    CmpGT,
    CmpEQ,
    CmpNE,

    LogicAnd,
    LogicXor,
    LogicOr,

    BitAnd,
    BitXor,
    BitOr,
    BitShiftLeft,
    BitShiftRight,
}

/// Unops.
#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub enum ASTUnop {
    LogicNot,
    BitNot,
    Len,
    Neg,
}

/// Expressions. For the partial equality we skip the span field to be able to test efficiently the
/// parsing.
#[derive(Debug)]
pub struct ASTExpr {
    pub content: ASTExprVariant,
    pub span: ASTSpan,
}

/// Expressions.
#[derive(Debug, PartialEq)]
pub enum ASTExprVariant {
    Nil,
    False,
    True,
    Integer(ASTInteger),
    Floating(ASTFloating),
    String(ASTString),
    Table(ASTTable<ASTExpr>),
    Binop(Box<ASTExpr>, ASTBinop, Box<ASTExpr>),
    Unop(ASTUnop, Box<ASTExpr>),
    FuncCall(Box<ASTExpr>, Vec<ASTExpr>),
    MethodInvok(Box<ASTExpr>, ASTString, Vec<ASTExpr>),
    PrefixExpr(Box<ASTExpr>, Vec<ASTField>),
    Tuple(Vec<ASTExpr>),
    Var(ASTVar),
    Const(ASTConst),
}

/// Fields indexes can be expressions or identifiers
#[derive(Debug, PartialEq)]
pub enum ASTField {
    Expr(ASTSpan, ASTExprVariant),
    Identifier(ASTSpan, ASTString),
}

/// Variable thing. Have a name, a where-it-is-defined span and optionally a type. Having to no
/// type means that the type was not already found.
#[derive(Debug, PartialEq, Eq)]
pub struct ASTVar(pub ASTSpan, pub ASTString, pub Option<ASTType>);

/// A constant expr.
#[derive(Debug, PartialEq)]
pub enum ASTConst {
    Color(ASTInteger),
    String(ASTString),
    Integer(ASTInteger),
    Floating(ASTFloating),
    Table(ASTTable<ASTConst>),
}

/// Types.
#[derive(Debug, PartialEq, Eq)]
pub enum ASTType {
    Integer,
    Floating,
    String,
    Color,
    Line,
    Syllabe,
    Table(ASTTable<ASTType>),
}

impl PartialEq for ASTExpr {
    fn eq(&self, other: &Self) -> bool {
        self.content == other.content
    }
}

impl PartialEq for ASTInstr {
    fn eq(&self, other: &Self) -> bool {
        self.content == other.content
    }
}

impl ASTVar {
    pub fn span(&self) -> ASTSpan {
        self.0
    }

    pub fn name(&self) -> &ASTString {
        &self.1
    }
}

impl std::fmt::Display for ASTVar {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_str(self.1.as_ref())
    }
}
