use crate::{ast::ASTSpan, parser::*};
pub(crate) use nom::error::{ContextError as NomContextError, ParseError as NomParseError};
use thiserror::Error;

#[macro_export]
macro_rules! nom_err_error {
    (vvs: $code: ident $args: tt) => {
        Err(nom::Err::Error($crate::parser::error::VVError::$code $args))
    };

    (vvs bare: $code: ident $args: tt) => {
        nom::Err::Error($crate::parser::error::VVError::$code $args)
    };

    (nom: $i: expr, $code: ident) => {
        Err(nom::Err::Error(VVError::from_error_kind(
            $i,
            nom::error::ErrorKind::$code,
        )))
    };

    (nom bare: $i: expr, $code: ident) => {
        VVError::from_error_kind($i, nom::error::ErrorKind::$code)
    };
}
pub(crate) use nom_err_error;

#[macro_export]
macro_rules! nom_err_failure {
    (vvs: $code: ident $args: tt) => {
        Err(nom::Err::Failure($crate::parser::error::VVError::$code $args))
    };

    (vvs bare: $code: ident $args: tt) => {
        nom::Err::Failure($crate::parser::error::VVError::$code $args)
    };

    (nom: $i: expr, $code: ident) => {
        Err(nom::Err::Failure(VVError::from_error_kind(
            $i,
            nom::error::ErrorKind::$code,
        )))
    };

    (nom bare: $i: expr, $code: ident) => {
        VVError::from_error_kind($i, nom::error::ErrorKind::$code)
    };
}
pub(crate) use nom_err_failure;

#[derive(Debug, Error, PartialEq, Eq)]
pub enum VVError {
    #[error("got an error from nom at {0}: {1:?}")]
    Nom(ASTSpan, nom::error::ErrorKind),

    #[error("got a nom error needed error at {0}: {1:?}")]
    NomNeeded(ASTSpan, nom::Needed),

    #[error("got multiple errors:\n{0:?}")]
    Multiple(Vec<VVError>),

    #[error("failed to parse an integer at {0}: {1}")]
    ParseIntError(ASTSpan, std::num::ParseIntError),

    #[error("failed to parse a floating number at {0}: {1}")]
    ParseFloatError(ASTSpan, std::num::ParseFloatError),

    #[error("failed to get tag `{1}` at {0}")]
    FailedToGetTag(ASTSpan, &'static str),

    #[error("not an identifier at {0}")]
    NotAnIdentifier(ASTSpan),

    #[error("table index `{1}` redefined at {0}")]
    TableIndexRedefined(ASTSpan, ASTString),

    #[error("option `{1}.{2}` at line {} redefined, first definition at {0}", .2.span())]
    OptionRedefined(ASTSpan, ASTString, ASTVar),

    #[error("error at {0}: {1}")]
    ErrorMsg(ASTSpan, String),
}

pub type VVParserError<'a, T> = IResult<PSpan<'a>, T, VVError>;
pub type VVResult<T> = Result<T, VVError>;

impl<'a> NomContextError<PSpan<'a>> for VVError {}

impl<'a> NomParseError<PSpan<'a>> for VVError {
    fn from_error_kind(input: PSpan, kind: nom::error::ErrorKind) -> Self {
        Self::Nom(input.into(), kind)
    }

    fn append(input: PSpan, kind: nom::error::ErrorKind, other: Self) -> Self {
        VVError::Multiple(match other {
            VVError::Multiple(mut other) => {
                other.push(Self::from_error_kind(input, kind));
                other
            }
            other => vec![other, Self::from_error_kind(input, kind)],
        })
    }
}

impl<'a> nom::error::FromExternalError<PSpan<'a>, std::num::ParseIntError> for VVError {
    fn from_external_error(input: PSpan<'a>, _: nom::error::ErrorKind, e: std::num::ParseIntError) -> Self {
        Self::ParseIntError(input.into(), e)
    }
}

impl<'a> nom::error::FromExternalError<PSpan<'a>, std::num::ParseFloatError> for VVError {
    fn from_external_error(input: PSpan<'a>, _: nom::error::ErrorKind, e: std::num::ParseFloatError) -> Self {
        Self::ParseFloatError(input.into(), e)
    }
}

impl<'a> nom::error::FromExternalError<PSpan<'a>, nom::Err<VVError>> for VVError {
    fn from_external_error(i: PSpan, _: nom::error::ErrorKind, e: nom::Err<VVError>) -> Self {
        match e {
            nom::Err::Incomplete(e) => Self::NomNeeded(i.into(), e),
            nom::Err::Error(e) | nom::Err::Failure(e) => e,
        }
    }
}
