//! Module responsible to parse the user's code, composed of .VVL files where the jobs are defined
//! and .VVS files where the global workflow is written explicitly. Some parsing functions are
//! common between the two parser, those things are expressed here.

mod error;
mod string;
mod types;

pub mod vvl;
pub mod vvs;

use crate::ast::*;

pub(crate) use crate::parser::{error::*, string::string, types::types};
pub(crate) use nom::{
    branch::alt,
    bytes::complete::{is_not, tag, take_while, take_while_m_n},
    character::{
        complete::{alpha1, alphanumeric1, char, digit1, multispace0, multispace1},
        is_alphanumeric,
    },
    combinator::{cut, map, map_opt, map_res, opt, peek, recognize, value, verify},
    error::context,
    multi::{fold_many0, many0, many0_count, separated_list0, separated_list1},
    sequence::{delimited, pair, preceded, separated_pair, terminated, tuple},
    IResult, Parser as NomParser,
};
pub(crate) use nom_locate::LocatedSpan;
pub(crate) use std::{cell::RefCell, rc::Rc};

/// A parser span. Will be converted into an AST span latter.
type PSpan<'a> = LocatedSpan<&'a str>;

pub(crate) fn with_span<'a, T>(
    mut cb: impl NomParser<PSpan<'a>, T, VVError>,
) -> impl FnMut(PSpan<'a>) -> VVParserError<'a, (PSpan<'a>, T)> {
    move |input: PSpan<'a>| {
        let (input, _) = multispace0(input)?;
        let (i, res) = cb.parse(input)?;
        Ok((i, (input, res)))
    }
}

pub(crate) fn spaced_tag<'a>(str: &'static str) -> impl FnMut(PSpan<'a>) -> VVParserError<'a, ()> {
    delimited(multispace0, map(tag(str), |_| ()), multispace0)
}

pub(crate) fn keyword<'a>(kw: &'static str) -> impl FnMut(PSpan<'a>) -> VVParserError<'a, ()> {
    delimited(multispace0, plain_tag(kw), multispace0)
}

pub(crate) fn plain_tag<'a>(t: &'static str) -> impl FnMut(PSpan<'a>) -> VVParserError<'a, ()> {
    move |i| {
        let (input, remain) = take_while(|x| is_alphanumeric(x as u8))(tag(t)(i)?.0)?;
        remain
            .is_empty()
            .then_some((input, ()))
            .ok_or_else(|| nom_err_error!(vvs bare: FailedToGetTag (i.into(), t)))
    }
}

pub(crate) fn number(i: PSpan) -> VVParserError<ASTInteger> {
    alt((
        map_res(preceded(opt(tag("+")), digit1), |digit_str: PSpan| {
            digit_str.parse::<ASTInteger>()
        }),
        map_res(preceded(tag("-"), digit1), |digit_str: PSpan| {
            digit_str.parse::<ASTInteger>().map(|x| -x)
        }),
    ))(i)
}

pub(crate) fn float(i: PSpan) -> VVParserError<ASTFloating> {
    map(
        pair(
            opt(alt((tag("+"), tag("-")))),
            map_res(recognize(tuple((digit1, tag("."), digit1))), |f: PSpan| {
                f.fragment().parse::<f32>()
            }),
        ),
        |(sign, value)| {
            sign.map(|s| if s.starts_with('-') { -value } else { value })
                .unwrap_or(value)
        },
    )(i)
}

pub(crate) fn identifier(input: PSpan) -> VVParserError<&str> {
    let (input, _) = multispace0(input)?;
    let (i, id) = match recognize(pair(
        alt((alpha1::<PSpan, VVError>, tag("_"))),
        many0_count(alt((alphanumeric1, tag("_")))),
    ))(input)
    {
        Ok(ok) => ok,
        Err(_) => return nom_err_error!(vvs: NotAnIdentifier (input.into())),
    };
    Ok((i, *id.fragment()))
}

#[test]
fn test_floats_and_numbers() {
    assert_eq!(number("1".into()).unwrap().1, 1);
    assert_eq!(number("0".into()).unwrap().1, 0);
    assert_eq!(number("-1".into()).unwrap().1, -1);
    assert_eq!(float("1.0".into()).unwrap().1, 1.0);
    assert_eq!(float("0.0".into()).unwrap().1, 0.0);
    assert_eq!(float("-1.0".into()).unwrap().1, -1.0);
    assert!(float("1.".into()).is_err());
}
