use crate::{ast::*, instruction, parser::vvl::*};

pub(crate) fn do_end_block<'a, 'b>(
    r#do: &'static str,
    end: &'static str,
    cache: Rc<RefCell<ASTStringCacheHandle<'b>>>,
) -> impl Fn(PSpan<'a>) -> VVParserError<'a, Vec<ASTInstr>> + 'b
where
    'b: 'a,
{
    move |input| {
        let (input, _) = multispace0(input)?;
        delimited(keyword(r#do), many0(instruction(cache.clone())), keyword(end))(input)
    }
}

#[allow(dead_code)]
pub(crate) fn instruction<'a, 'b>(
    cache: Rc<RefCell<ASTStringCacheHandle<'b>>>,
) -> impl Fn(PSpan<'a>) -> VVParserError<'a, ASTInstr> + 'b
where
    'b: 'a,
{
    move |input| {
        let (input, _) = multispace0(input)?;
        terminated(
            alt((
                // Break/Continue loops
                map(with_span(keyword("break")), |(sp, _)| instruction!(sp, Break)),
                map(with_span(keyword("continue")), |(sp, _)| instruction!(sp, Continue)),
                // Variable assignation. Don't do the destructuring for now.
                map(
                    with_span(separated_pair(
                        expr(cache.clone()),
                        spaced_tag("="),
                        expr(cache.clone()),
                    )),
                    |(sp, (dest, src))| instruction!(sp, Assign(vec![dest], vec![src])),
                ),
                // Variable declaration. Don't do the destructuring for now.
                map(
                    with_span(tuple((
                        keyword("let"),
                        identifier,
                        opt(types),
                        spaced_tag("="),
                        expr(cache.clone()),
                    ))),
                    |(sp, (_, name, ty, _, decl))| {
                        let var = ASTVar(sp.into(), cache.borrow_mut().get_or_insert(name), ty);
                        instruction!(sp, Decl(vec![var], vec![decl]))
                    },
                ),
                // A do-end block with a list of instructions inside it.
                map(with_span(do_end_block("do", "end", cache.clone())), |(sp, body)| {
                    instruction!(sp, Block(body))
                }),
                // While loop.
                map(
                    with_span(pair(
                        preceded(keyword("while"), expr(cache.clone())),
                        do_end_block("do", "end", cache.clone()),
                    )),
                    |(sp, (cond, body))| instruction!(sp, WhileDo(cond, body)),
                ),
                // A DoWhile block (or a reapeat until in lua...)
                map(
                    with_span(pair(do_end_block("do", "until", cache.clone()), expr(cache.clone()))),
                    |(sp, (body, cond))| instruction!(sp, RepeatUntil(cond, body)),
                ),
            )),
            // We may terminate with a semicolon
            opt(spaced_tag(";")),
        )(input)
    }
}
