use super::*;
use hashbrown::HashMap;

/// Parses the keys in tables or returns an error. Keys can be identifiers or a string between
/// square braquets: `[`, `]`.
fn key<'a, 'b>(cache: Rc<RefCell<ASTStringCacheHandle<'b>>>) -> impl Fn(PSpan<'a>) -> VVParserError<'a, ASTString> + 'b
where
    'b: 'a,
{
    move |input: PSpan| -> VVParserError<ASTString> {
        let (input, _) = multispace0(input)?;
        alt((
            map(delimited(spaced_tag("["), string, spaced_tag("]")), |str| {
                cache.borrow_mut().get_or_insert(str)
            }),
            map(identifier, |str| cache.borrow_mut().get_or_insert(str)),
        ))(input)
    }
}

/// Parses a separator for key value pairs for the table parser. In Lua separators can be `,` or
/// `;` for some reasons...
fn separator(input: PSpan) -> VVParserError<()> {
    alt((spaced_tag(","), spaced_tag(";")))(input)
}

/// Parses a table or returns an error.
pub(crate) fn table<'a, 'b>(
    cache: Rc<RefCell<ASTStringCacheHandle<'b>>>,
) -> impl Fn(PSpan<'a>) -> VVParserError<'a, ASTTable<ASTExpr>> + 'b
where
    'b: 'a,
{
    move |input| {
        let (input, _) = multispace0(input)?;
        let table_key_value_pair = separated_pair(key(cache.clone()), spaced_tag("="), expr(cache.clone()));
        let (i, content) = context(
            "table",
            delimited(
                spaced_tag("{"),
                separated_list0(separator, table_key_value_pair),
                preceded(opt(separator), spaced_tag("}")),
            ),
        )(input)?;
        let hashmap = HashMap::with_capacity(content.len());
        let hashmap = content.into_iter().try_fold(hashmap, |mut hashmap, (key, value)| {
            if hashmap.contains_key(&key) {
                nom_err_error!(vvs: TableIndexRedefined (value.span, key))
            } else {
                hashmap.insert(key, value);
                Ok(hashmap)
            }
        })?;
        Ok((i, hashmap))
    }
}
