use crate::StringSlice;
use std::{
    ffi::{c_char, CStr},
    ptr::NonNull,
};
use vvs_ass::{ass_container_from_file, ASSContainer, ASSLine, ASSStyle, ASSSyllabe};

/// Contains wrappers for the styles.
pub mod style {
    use super::*;

    /// Get the name of the style.
    ///
    /// # Safety
    /// It is up to the user to ensure that the style is not dropped before the returned pointer.
    #[no_mangle]
    pub extern "C" fn ASSStyleGetName(this: &ASSStyle) -> StringSlice {
        StringSlice::from(&this.name)
    }
}

/// Contains wrappers for the container, the line, syllabes, etc...
pub mod elements {
    use super::*;

    /// Get the number of lines in the container.
    #[no_mangle]
    pub extern "C" fn ASSContainerGetLinesCount(this: &ASSContainer) -> usize {
        this.lines.len()
    }

    /// Get the number of syllabes in the line.
    #[no_mangle]
    pub extern "C" fn ASSLineGetSyllabesCount(this: &ASSLine) -> usize {
        this.content.len()
    }

    /// # Safety
    /// It is up to the user to ensure that no aliasing is done or that the line is not removed
    /// from the container before the pointer is dropped, or that the container is not dropped...
    #[no_mangle]
    pub unsafe extern "C" fn ASSContainerGetLineAt(this: &mut ASSContainer, idx: usize) -> *mut ASSLine {
        match this.lines.get(idx) {
            Some(line) => match line.0.write() {
                Ok(mut line) => (&mut *line) as *mut _,
                Err(err) => {
                    log::error!("failed to get line n°{idx}: {err}");
                    std::ptr::null_mut()
                }
            },
            None => {
                log::error!("failed to get line n°{idx}");
                std::ptr::null_mut()
            }
        }
    }

    /// # Safety
    /// It is up to the user to ensure that no aliasing is done or that the syllabe is not removed
    /// from the line before the pointer is dropped, or that the line is not removed from the
    /// container...
    #[no_mangle]
    pub unsafe extern "C" fn ASSLineGetSyllabeAt(this: &mut ASSLine, idx: usize) -> *mut ASSSyllabe {
        match this.content.get(idx) {
            Some(s) => match s.0.write() {
                Ok(mut s) => (&mut *s) as *mut _,
                Err(err) => {
                    log::error!("failed to get syllabe n°{idx}: {err}");
                    std::ptr::null_mut()
                }
            },
            None => {
                log::error!("failed to get syllabe n°{idx}");
                std::ptr::null_mut()
            }
        }
    }

    /// Tells whever a line is commented or not.
    #[no_mangle]
    pub extern "C" fn ASSLineIsCommented(this: &ASSLine) -> bool {
        this.is_comment
    }

    /// Gets the style of the line, or 'Default' if it was not found/specified.
    #[no_mangle]
    pub extern "C" fn ASSLineGetStyle(this: &ASSLine) -> StringSlice {
        this.aux
            .get("style")
            .map(|style| match style {
                vvs_ass::ASSAuxValue::String(str) => StringSlice::from(str),
                _ => StringSlice::from("Default"),
            })
            .unwrap_or_else(|| StringSlice::from("Default"))
    }

    /// # Safety
    /// The returned pointer must be freed
    #[no_mangle]
    pub extern "C" fn ASSSyllabeGetContent(this: &ASSSyllabe) -> StringSlice {
        StringSlice::from(&this.content)
    }

    #[no_mangle]
    pub extern "C" fn ASSSyllabeGetDuration(this: &ASSSyllabe) -> i64 {
        this.fini.saturating_sub(this.start)
    }

    #[no_mangle]
    pub extern "C" fn ASSSyllabeSetDuration(this: &mut ASSSyllabe, duration: i64) {
        this.fini = this.start.saturating_add(duration);
    }

    /// Load the ASS from a file, returns nullptr if an error was encountred.
    #[no_mangle]
    pub extern "C" fn ASSContainerFromFile(path: NonNull<c_char>) -> *mut ASSContainer {
        match unsafe { CStr::from_ptr(path.as_ptr() as *const _) }.to_str() {
            Ok(path) => ass_container_from_file(path)
                .map(|ass| Box::into_raw(Box::new(ass)))
                .unwrap_or_else(|err| {
                    log::error!("failed to load file: {err}");
                    std::ptr::null_mut()
                }),
            Err(err) => {
                log::error!("invalid utf8 filename: {err}");
                return std::ptr::null_mut();
            }
        }
    }

    /// # Safety
    /// The name must be a valid CString (null terminated.)
    #[no_mangle]
    pub extern "C" fn ASSContainerGetStyleByName(this: &mut ASSContainer, name: StringSlice) -> *mut ASSStyle {
        let name = unsafe { name.as_str() };
        match this.styles.get_mut(name) {
            Some(style) => style as *mut _,
            None => {
                log::error!("no sytle {name} in ass container");
                std::ptr::null_mut()
            }
        }
    }

    #[no_mangle]
    pub extern "C" fn ASSContainerGetStyleNameAt(this: &ASSContainer, idx: usize) -> StringSlice {
        let mut keys: Vec<_> = this.styles.keys().collect();
        keys.sort();
        keys.get(idx)
            .map(StringSlice::from)
            .unwrap_or_else(|| StringSlice::from("Default"))
    }

    #[no_mangle]
    pub extern "C" fn ASSContainerGetStylesCount(this: &ASSContainer) -> usize {
        this.styles.len()
    }

    #[no_mangle]
    pub extern "C" fn ASSLineSetStartTime(this: &mut ASSLine, time: i64) {
        this.start = time;
        this.fini = this.fini.max(time);
        for syl in &this.content {
            match syl.0.write() {
                Ok(mut syl) => {
                    syl.start = syl.start.max(time);
                    syl.fini = syl.fini.max(time);
                }
                Err(err) => log::error!("{err}"),
            }
        }
    }

    #[no_mangle]
    pub extern "C" fn ASSLineSetFiniTime(this: &mut ASSLine, time: i64) {
        this.fini = time;
        this.start = this.start.min(time);
        for syl in &this.content {
            match syl.0.write() {
                Ok(mut syl) => {
                    syl.start = syl.start.min(time);
                    syl.fini = syl.fini.min(time);
                }
                Err(err) => log::error!("{err}"),
            }
        }
    }

    #[no_mangle]
    pub extern "C" fn ASSLineGetDuration(this: &ASSLine) -> i64 {
        this.fini.saturating_sub(this.start)
    }
    #[no_mangle]
    pub extern "C" fn ASSLineGetStartTime(this: &ASSLine) -> i64 {
        this.start
    }

    #[no_mangle]
    pub extern "C" fn ASSLineGetFiniTime(this: &ASSLine) -> i64 {
        this.fini
    }

    /// Drop the container. It is file to pass a null pointer to this function.
    /// # Safety
    /// The use must ensure that no dangling references remains...
    #[no_mangle]
    pub extern "C" fn ASSContainerDrop(this: *mut ASSContainer) {
        if !this.is_null() {
            drop(unsafe { Box::from_raw(this) });
        }
    }
}
