use crate::xdg::*;
use std::{
    collections::hash_map::DefaultHasher,
    fs::{File, OpenOptions},
    hash::{Hash, Hasher},
    io::Error as IoError,
    path::Path,
};
use thiserror::Error;

/// An error can be comming from the IO thing or the resource could be already locked.
#[derive(Debug, Error)]
pub enum LockError {
    #[error("{0}")]
    IO(std::io::Error),

    #[error("resource is already locked")]
    AlreadyLocked,

    #[error("{0}")]
    XDG(#[from] XDGError),
}

/// Result type for lock functions.
pub type LockResult<T> = Result<T, LockError>;

/// Structure used to hold the lock file for a resource that can be hashed.
pub struct LockFile(File);

impl LockFile {
    fn handle_xdg_error(err: XDGError) -> LockError {
        match err {
            XDGError::ConfigIO(_, _, err) => Self::handle_io_error(err),
            err => LockError::XDG(err),
        }
    }

    fn handle_io_error(err: IoError) -> LockError {
        match err.kind() {
            std::io::ErrorKind::AlreadyExists => LockError::AlreadyLocked,
            _ => LockError::IO(err),
        }
    }

    fn create(app: impl AsRef<str>, file: String) -> LockResult<Self> {
        match XDGFolder::RuntimeDir
            .find(app, file, XDGFindBehaviour::NonExistingOnly)
            .map_err(Self::handle_xdg_error)?
        {
            MaybeFolderList::Empty => unreachable!("failed to find the runtime dir..."),
            MaybeFolderList::Folder(f) | MaybeFolderList::Many(f, _) => Ok(Self(
                OpenOptions::new()
                    .create_new(true)
                    .open(f)
                    .map_err(Self::handle_io_error)?,
            )),
        }
    }

    /// Create a lock file for another file.
    pub fn for_file(app: impl AsRef<str>, path: impl AsRef<Path>) -> LockResult<Self> {
        let mut s = DefaultHasher::new();
        path.as_ref().hash(&mut s);
        Self::create(app, format!("{:#X}.path.lock", s.finish()))
    }

    /// Create a lock file for an hashable ressource.
    pub fn for_resource(app: impl AsRef<str>, rsc: impl Hash) -> LockResult<Self> {
        let mut s = DefaultHasher::new();
        rsc.hash(&mut s);
        Self::create(app, format!("{:#X}.rsc.lock", s.finish()))
    }

    /// Create a lock file for an ID.
    pub fn for_id(app: impl AsRef<str>, id: u64) -> LockResult<Self> {
        Self::create(app, format!("{id:#X}.id.lock"))
    }
}

impl Drop for LockFile {
    fn drop(&mut self) {
        todo!("drop {:?}", self.0)
    }
}
