//! Simple Mersenne Twister (MT19937-64) implementation.

use crate::either;
use std::{cell::RefCell, sync::OnceLock};

/// Simple Mersenne Twister (MT19937-64) implementation.
struct MersenneTwister {
    mt: [u64; MersenneTwister::W as usize],
    index: usize,
}

thread_local! {
    /// We do our things here.
    static MERSENNE: OnceLock<RefCell<MersenneTwister>> = OnceLock::new();
}

/// Get the seed for the thread.
fn get_seed() -> u64 {
    f64::to_bits(
        std::time::SystemTime::now()
            .duration_since(std::time::UNIX_EPOCH)
            .expect("time leap before the UNIX_EPOCH")
            .as_secs_f64(),
    )
}

impl MersenneTwister {
    const W: u64 = u64::BITS as u64;
    const N: u64 = 624;
    const M: u64 = 397;
    const R: u64 = 31;

    const A: u64 = 0xb5026f5aa96619e9;
    const B: u64 = 0x71d67fffeda60000;
    const C: u64 = 0xfff7eee000000000;
    const D: u64 = 0x5555555555555555;
    const S: u64 = 17;
    const T: u64 = 37;
    const U: u64 = 29;
    const L: u64 = 43;
    const F: u64 = 6364136223846793005;

    const LOWER_MASK: u64 = (1_u64 << MersenneTwister::R).overflowing_sub(1_u64).0;
    const UPPER_MASK: u64 = !Self::LOWER_MASK;

    fn twist(&mut self) {
        (0..Self::N).for_each(|i| {
            let index = i as usize;
            let index_next = (i + 1) as usize % Self::W as usize;
            let x = (self.mt[index] & Self::UPPER_MASK) | (self.mt[index_next] & Self::LOWER_MASK);
            self.mt[index] =
                self.mt[((i + Self::M) % Self::N) as usize] ^ either!((x % 2) != 0 => (x >> 1) ^ Self::A; x >> 1);
        });
        self.index = 0;
    }

    fn extract_number(&mut self) -> u64 {
        if self.index >= Self::N as usize {
            self.twist();
        }
        let y = self.mt[self.index];
        let y = y ^ ((y >> Self::U) & Self::D);
        let y = y ^ ((y << Self::S) & Self::B);
        let y = y ^ ((y << Self::T) & Self::C);
        let y = y ^ (y >> Self::L);
        self.index += 1;
        y
    }
}

/// Get next pseudo random number with our homebrew mersenne twister.
pub fn rand() -> u64 {
    MERSENNE.with(|mt| {
        mt.get_or_init(|| {
            let index = MersenneTwister::N as usize;
            let mut mt = [0; MersenneTwister::W as usize];
            mt[0] = get_seed();
            (1..MersenneTwister::N).for_each(|i| {
                let index = i as usize - 1;
                let factor = (mt[index] ^ (MersenneTwister::F * (mt[index] >> (MersenneTwister::W - 2)))) + i;
                mt[index] = MersenneTwister::F.overflowing_mul(factor).0;
            });
            RefCell::new(MersenneTwister { mt, index })
        })
        .borrow_mut()
        .extract_number()
    })
}
