//! The main provider for resolving files with the XDG specification + some utilities.

use crate::{
    either,
    xdg::{MaybeFolderList, XDGError, XDGFindBehaviour},
};
use std::{io::ErrorKind as IoErrorKind, path::PathBuf};

/// The type of folder we want. Here are some remarks from the specification about file resolution:
///
/// - If, when attempting to write a file, the destination directory is non-existent an attempt
///   should be made to create it with permission 0700. If the destination directory exists already
///   the permissions should not be changed. The application should be prepared to handle the case
///   where the file could not be written, either because the directory was non-existent and could
///   not be created, or for any other reason. In such case it may choose to present an error
///   message to the user.
/// - When attempting to read a file, if for any reason a file in a certain directory is
///   unaccessible, e.g. because the directory is non-existent, the file is non-existent or the
///   user is not authorized to open the file, then the processing of the file in that directory
///   should be skipped. If due to this a required file could not be found at all, the application
///   may choose to present an error message to the user.
/// - A specification that refers to $XDG_DATA_DIRS or $XDG_CONFIG_DIRS should define what the
///   behaviour must be when a file is located under multiple base directories. It could, for
///   example, define that only the file under the most important base directory should be used or,
///   as another example, it could define rules for merging the information from the different
///   files.
#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub enum XDGFolder {
    /// $XDG_DATA_HOME defines the base directory relative to which user-specific data files should
    /// be stored. If $XDG_DATA_HOME is either not set or empty, a default equal to
    /// $HOME/.local/share should be used.
    DataHome,

    /// $XDG_CONFIG_HOME defines the base directory relative to which user-specific configuration
    /// files should be stored. If $XDG_CONFIG_HOME is either not set or empty, a default equal to
    /// $HOME/.config should be used.
    ConfigHome,

    /// $XDG_STATE_HOME defines the base directory relative to which user-specific state files
    /// should be stored. If $XDG_STATE_HOME is either not set or empty, a default equal to
    /// $HOME/.local/state should be used.
    ///
    /// The $XDG_STATE_HOME contains state data that should persist between (application) restarts,
    /// but that is not important or portable enough to the user that it should be stored in
    /// $XDG_DATA_HOME. It may contain:
    /// - actions history (logs, history, recently used files, …)
    /// - current state of the application that can be reused on a restart (view, layout, open
    ///   files, undo history, …)
    StateHome,

    /// $XDG_DATA_DIRS defines the preference-ordered set of base directories to search for data
    /// files in addition to the $XDG_DATA_HOME base directory. The directories in $XDG_DATA_DIRS
    /// should be seperated with a colon ':'.
    ///
    /// If $XDG_DATA_DIRS is either not set or empty, a value equal to
    /// /usr/local/share/:/usr/share/ should be used.
    DataDirs,

    /// $XDG_CONFIG_DIRS defines the preference-ordered set of base directories to search for
    /// configuration files in addition to the $XDG_CONFIG_HOME base directory. The directories in
    /// $XDG_CONFIG_DIRS should be seperated with a colon ':'.
    ///
    /// If $XDG_CONFIG_DIRS is either not set or empty, a value equal to /etc/xdg should be used.
    ConfigDirs,

    /// $XDG_CACHE_HOME defines the base directory relative to which user-specific non-essential
    /// data files should be stored. If $XDG_CACHE_HOME is either not set or empty, a default equal
    /// to $HOME/.cache should be used.
    CacheHome,

    /// $XDG_RUNTIME_DIR defines the base directory relative to which user-specific non-essential
    /// runtime files and other file objects (such as sockets, named pipes, ...) should be stored.
    /// The directory MUST be owned by the user, and he MUST be the only one having read and write
    /// access to it. Its Unix access mode MUST be 0700.
    ///
    /// The lifetime of the directory MUST be bound to the user being logged in. It MUST be created
    /// when the user first logs in and if the user fully logs out the directory MUST be removed.
    /// If the user logs in more than once he should get pointed to the same directory, and it is
    /// mandatory that the directory continues to exist from his first login to his last logout on
    /// the system, and not removed in between. Files in the directory MUST not survive reboot or a
    /// full logout/login cycle.
    ///
    /// The directory MUST be on a local file system and not shared with any other system. The
    /// directory MUST by fully-featured by the standards of the operating system. More
    /// specifically, on Unix-like operating systems AF_UNIX sockets, symbolic links, hard links,
    /// proper permissions, file locking, sparse files, memory mapping, file change notifications,
    /// a reliable hard link count must be supported, and no restrictions on the file name
    /// character set should be imposed. Files in this directory MAY be subjected to periodic
    /// clean-up. To ensure that your files are not removed, they should have their access time
    /// timestamp modified at least once every 6 hours of monotonic time or the 'sticky' bit should
    /// be set on the file.
    ///
    /// If $XDG_RUNTIME_DIR is not set applications should fall back to a replacement directory
    /// with similar capabilities and print a warning message. Applications should use this
    /// directory for communication and synchronization purposes and should not place larger files
    /// in it, since it might reside in runtime memory and cannot necessarily be swapped out to
    /// disk.
    RuntimeDir,
}

impl XDGFolder {
    /// The list separator.
    const SEPARATOR: char = ':';

    /// Get the folders or folder list that the user asked for. If the env variable is not present,
    /// falls back to the default value.
    ///
    /// Note that the folder or folders doesn't exist, this function won't try to create them.
    pub fn get_folder(&self) -> MaybeFolderList {
        #[cfg(unix)]
        mod variables {
            pub fn data_dirs(_: impl AsRef<std::path::Path>) -> super::MaybeFolderList {
                super::MaybeFolderList::from_iter(["/usr/local/share", "/usr/share"])
            }
            pub fn config_dirs(_: impl AsRef<std::path::Path>) -> super::MaybeFolderList {
                super::MaybeFolderList::from("/etc/xdg")
            }
        }
        #[cfg(not(unix))]
        mod variables {
            pub fn data_dirs(_: impl AsRef<std::path::Path>) -> super::MaybeFolderList {
                super::MaybeFolderList::Folder(home.join(".local/share"))
            }
            pub fn config_dirs(_: impl AsRef<std::path::Path>) -> super::MaybeFolderList {
                super::MaybeFolderList::Folder(home.join(".config"))
            }
        }

        let home = crate::xdg::home_folder();
        match std::env::var(self.env_var_name()) {
            Ok(folder_list) if self.is_list() => MaybeFolderList::from_str_list(&folder_list, Self::SEPARATOR),
            Ok(folder) => MaybeFolderList::Folder(PathBuf::from(folder)),
            Err(_) => match self {
                XDGFolder::DataHome => MaybeFolderList::Folder(home.join(".local/share")),
                XDGFolder::ConfigHome => MaybeFolderList::Folder(home.join(".config")),
                XDGFolder::StateHome => MaybeFolderList::Folder(home.join(".local/state")),
                XDGFolder::CacheHome => MaybeFolderList::Folder(home.join(".cache")),
                XDGFolder::RuntimeDir => panic!("failed to find the env variable $XDG_RUNTIME_DIR"),
                XDGFolder::DataDirs => variables::data_dirs(&home),
                XDGFolder::ConfigDirs => variables::config_dirs(&home),
            },
        }
    }

    /// Get the folders list of what you asked for, but try to create them and only return existing
    /// ones. The thing returned is a thing that can be iterate over to facilitate chaining.
    pub fn prepare_folder(&self) -> impl IntoIterator<Item = <MaybeFolderList as IntoIterator>::Item> {
        self.get_folder().into_iter().filter_map(|folder| {
            if folder.is_dir() {
                Some(folder)
            } else {
                std::fs::create_dir_all(&folder)
                    .map(|()| folder)
                    .map_err(|err| log::error!(target: "xdg", "{err}"))
                    .ok()
            }
        })
    }

    /// Get the file associated with the said application in the specified folders. An option can
    /// be passed to specify behaviour in the search or file/folder creation processes.
    pub fn find(
        &self,
        app: impl AsRef<str>,
        file: impl AsRef<str>,
        opt: XDGFindBehaviour,
    ) -> Result<MaybeFolderList, XDGError> {
        enum FindState {
            CanCreate(PathBuf),
            Exists(PathBuf),
        }
        impl<'a> std::iter::Extend<&'a FindState> for MaybeFolderList {
            fn extend<T: IntoIterator<Item = &'a FindState>>(&mut self, iter: T) {
                self.extend(iter.into_iter().map(|m| match m {
                    FindState::CanCreate(path) | FindState::Exists(path) => path.clone(),
                }));
            }
        }

        // DataHome is more important than DataDirs in the search process... Idem for the Config
        // env var family...
        let matches: Vec<_> = match self {
            XDGFolder::DataDirs => XDGFolder::DataHome.get_folder(),
            XDGFolder::ConfigDirs => XDGFolder::ConfigHome.get_folder(),
            _ => MaybeFolderList::Empty,
        }
        .into_iter()
        .chain(self.get_folder())
        .flat_map(|path| {
            let path = path.join(app.as_ref());
            if let Err(err) = std::fs::create_dir_all(&path) {
                if !matches!(err.kind(), IoErrorKind::PermissionDenied | IoErrorKind::AlreadyExists) {
                    log::error!(target: "xdg",
                        "failed to create app folder for {{{self}}}/{}/{} as {}",
                        app.as_ref(), file.as_ref(), path.to_string_lossy()
                    );
                    return None;
                }
            }
            let path = path.join(file.as_ref());
            Some(either!(path.exists() => FindState::Exists(path); FindState::CanCreate(path)))
        })
        .collect();

        // For now, default behaviour is to return the first file found or try to create it, the
        // first file created will be returned (if we can't create the file in system folders it's
        // not an error...)

        let (mut exists, mut can_create): (MaybeFolderList, MaybeFolderList) =
            matches.iter().partition(|m| matches!(m, FindState::Exists(_)));
        use XDGFindBehaviour::*;
        match (exists.is_some(), can_create.is_some()) {
            // [XDGFindBehaviour::FirstOrCreate] case...
            (true, false) if opt == FirstOrCreate => Ok(exists.into_first().into_iter().collect()),
            (false, true) if opt == FirstOrCreate => Ok(can_create.into_first().into_iter().collect()),

            // Should we return the one list that is not empty?
            (true, false) if matches!(opt, AllFiles | ExistingOnly) => Ok(exists),
            (false, true) if matches!(opt, AllFiles | NonExistingOnly) => Ok(can_create),

            // Complicated case
            (true, true) => match opt {
                AllFiles => {
                    exists.append(&mut can_create);
                    Ok(exists)
                }

                FirstOrCreate => Ok(exists.into_first().into_iter().collect()),
                ExistingOnly => Ok(exists),
                NonExistingOnly => Ok(can_create),
            },

            // The list that we want is empty, or all lists are empty, returns nothing -> should it
            // really be an error?
            _ => Err(XDGError::NotFound(
                *self,
                app.as_ref().to_string(),
                file.as_ref().to_string(),
            )),
        }
    }

    /// Is this variable a folder list?
    pub fn is_list(&self) -> bool {
        use XDGFolder::*;
        matches!(self, DataDirs | ConfigDirs)
    }

    /// Get the env variable name associated with the folder or folder list.
    pub fn env_var_name(&self) -> &str {
        self.as_ref()
    }
}

impl AsRef<str> for XDGFolder {
    fn as_ref(&self) -> &str {
        match self {
            XDGFolder::DataHome => "XDG_DATA_HOME",
            XDGFolder::StateHome => "XDG_STATE_HOME",
            XDGFolder::CacheHome => "XDG_CACHE_HOME",
            XDGFolder::ConfigHome => "XDG_CONFIG_HOME",

            XDGFolder::DataDirs => "XDG_DATA_DIRS",
            XDGFolder::ConfigDirs => "XDG_CONFIG_DIRS",
            XDGFolder::RuntimeDir => "XDG_RUNTIME_DIR",
        }
    }
}

impl std::fmt::Display for XDGFolder {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "${}", self.as_ref())
    }
}
