//! A PathBuf container as a sum type.

use std::{
    mem,
    path::{Path, PathBuf},
};

/// An enum to store maybe multiple folders, or none, with utility functions.
#[derive(Debug, Default, Clone, PartialEq, Eq)]
pub enum MaybeFolderList {
    /// No folder.
    #[default]
    Empty,

    /// Only one folder.
    Folder(PathBuf),

    /// Many folders.
    Many(PathBuf, Vec<PathBuf>),
}

/// An interator over [MaybeFolderList].
#[derive(Debug, Clone, PartialEq, Eq)]
pub struct MaybeFolderListIterator<'a> {
    next: Option<&'a PathBuf>,
    list: &'a [PathBuf],
}

impl MaybeFolderList {
    /// Is there no folder?
    pub fn is_none(&self) -> bool {
        matches!(self, MaybeFolderList::Empty)
    }

    /// Is there folders?
    pub fn is_some(&self) -> bool {
        !self.is_none()
    }

    /// Get the first folder, or the last folder. If no folder is present just returns [None].
    pub fn first(&self) -> Option<&Path> {
        use MaybeFolderList::*;
        match self {
            Empty => None,
            Folder(f) | Many(f, _) => Some(f.as_ref()),
        }
    }

    /// Append another [MaybeFolderList] at the end of a [MaybeFolderList]. Empties the other
    /// variable and try to reuse memory.
    pub fn append(&mut self, other: &mut MaybeFolderList) {
        use MaybeFolderList::*;

        match (self, other) {
            (this @ Empty, other) => match other {
                Empty => {}
                Folder(f1) => {
                    *this = Folder(mem::take(f1));
                    *other = Empty;
                }
                Many(f1, fs) => {
                    *this = Many(mem::take(f1), mem::take(fs));
                    *other = Empty;
                }
            },

            (_, Empty) => {}

            (Many(_, fs), other @ Folder(_)) => {
                let Folder(f2) = other else { unreachable!() };
                fs.push(mem::take(f2));
                *other = Empty;
            }
            (Many(_, fs), other @ Many(_, _)) => {
                let Many(f2, ref mut fs2) = other else { unreachable!() };
                fs.reserve(fs2.len() + 1);
                fs.push(mem::take(f2));
                fs.append(fs2);
                *other = Empty;
            }

            (this @ Folder(_), other) => {
                let Folder(f1) = this else { unreachable!() };
                match other {
                    Empty => unreachable!(),
                    Folder(f2) => *this = Many(mem::take(f1), vec![mem::take(f2)]),
                    Many(f2, ref mut fs) => {
                        fs.insert(0, mem::take(f2));
                        *this = Many(mem::take(f1), mem::take(fs));
                    }
                }
                *other = Empty;
            }
        }
    }

    /// Get the first folder, or the last folder. If no folder is present just returns [None]. The
    /// difference from the [MaybeFolderList::first] function is that this function consumes the
    /// [MaybeFolderList].
    pub fn into_first(self) -> Option<PathBuf> {
        use MaybeFolderList::*;
        match self {
            Empty => None,
            Folder(f) | Many(f, _) => Some(f),
        }
    }

    /// Get an iterator over the contained folders.
    pub fn iter(&self) -> MaybeFolderListIterator {
        let (next, list): (_, &[_]) = match self {
            MaybeFolderList::Empty => (None, &[]),
            MaybeFolderList::Folder(f) => (Some(f), &[]),
            MaybeFolderList::Many(f, fs) => (Some(f), &fs[..]),
        };
        MaybeFolderListIterator { next, list }
    }

    /// Create a list of folders from a description string and a separator.
    pub(super) fn from_str_list(list: &str, separator: char) -> Self {
        let mut folders = list.split(separator);
        match folders.next() {
            None => MaybeFolderList::Empty,
            Some(first) => MaybeFolderList::Many(PathBuf::from(first), folders.map(PathBuf::from).collect()),
        }
    }
}

impl std::iter::Extend<PathBuf> for MaybeFolderList {
    fn extend<T: IntoIterator<Item = PathBuf>>(&mut self, iter: T) {
        use MaybeFolderList::*;
        let mut iter = iter.into_iter();
        match self {
            Empty => {
                if let Some(next) = iter.next() {
                    *self = Many(next, iter.collect());
                }
            }
            Folder(f1) => *self = Many(mem::take(f1), iter.collect()),
            Many(_, fs) => fs.extend(iter),
        }
    }
}

impl<'a> std::iter::Extend<&'a MaybeFolderList> for MaybeFolderList {
    fn extend<T: IntoIterator<Item = &'a MaybeFolderList>>(&mut self, iter: T) {
        use MaybeFolderList::*;
        let other = iter.into_iter().flat_map(|iter| iter.iter());
        match self {
            Empty => *self = other.collect(),
            Folder(f1) => *self = Many(mem::take(f1), other.cloned().collect()),
            Many(_, ref mut fs) => fs.extend(other.cloned()),
        }
    }
}

impl<'a, S> From<S> for MaybeFolderList
where
    S: AsRef<str> + 'a,
{
    fn from(value: S) -> Self {
        MaybeFolderList::Folder(PathBuf::from(value.as_ref()))
    }
}

impl<'a, P> FromIterator<P> for MaybeFolderList
where
    P: AsRef<Path> + 'a,
{
    fn from_iter<T: IntoIterator<Item = P>>(iter: T) -> Self {
        let mut folders = iter.into_iter();
        match folders.next() {
            None => MaybeFolderList::Empty,
            Some(first) => MaybeFolderList::Many(
                first.as_ref().to_path_buf(),
                folders.map(|p| p.as_ref().to_path_buf()).collect(),
            ),
        }
    }
}

impl IntoIterator for MaybeFolderList {
    type Item = PathBuf;
    type IntoIter = <Vec<PathBuf> as IntoIterator>::IntoIter;

    fn into_iter(self) -> Self::IntoIter {
        match self {
            MaybeFolderList::Empty => vec![].into_iter(),
            MaybeFolderList::Folder(f) => vec![f].into_iter(),
            MaybeFolderList::Many(f, mut fs) => {
                fs.insert(0, f);
                fs.into_iter()
            }
        }
    }
}

impl<'a> Iterator for MaybeFolderListIterator<'a> {
    type Item = &'a PathBuf;

    fn next(&mut self) -> Option<Self::Item> {
        let next = self.next?;
        match self.list.split_first() {
            Some((next, list)) => {
                self.next = Some(next);
                self.list = list;
            }
            None => self.next = None,
        }
        Some(next)
    }
}
