#include "PreCompiledHeaders.hh"
#include "UI/AboutWindow.hh"
#include "VivyApplication.hh"

using namespace Vivy;

// See https://doc.qt.io/qt-5/richtext-html-subset.html for the supported
// rich text subset.

static const char *aboutContent =
    "<body>"
    "  <p>Vivy is a replacement for Aegisub, writen in Qt6 and with less segfaults - hopefully.</p>"
    "  <p>This software (Vivy) is licenced under the LGPL v3.0 or latter and is &copy; Vivy contributors."
    "     The libvvs library and vvcc compiler are MIT licenced and are &copy; 2023 Vivy contributors.</p>"
    "  <p>Contributors:<ul>"
    "    <li>Elliu</li>"
    "    <li>Kubat</li>"
    "  </ul></p>"
    "</body>";

static const char *libContent =
    "<body>"
    "  <p>Here are the libraries that where used to create the software:</p>"
    "  <ul>"
    "    <li><a href=https://doc.qt.io/qt-6/index.html>Qt6</a> &copy; <i>2023 The Qt Company Ltd. and other contributors. (LGPL-V3)</i></li>"
    "    <li><a href=https://libav.org>Libav</a> &copy; <i>Libav contributors. (LGPL-V2.1+)</i></li>"
    "  </ul>"
    "  <p>Other external dependencies where used:</p>"
    "  <ul>"
    "    <li><a href=https://github.com/tonsky/FiraCode>FiraCode</a> &copy; <i>Nikita Prokopov. (OFL-1.1)</i></li>"
    "    <li><a href=https://fonts.google.com/specimen/Noto+Sans>NotoSans</a> &copy; <i>Google. (APACHE-2.0)</i>"
    "    <li><a href=https://github.com/KDE/breeze-icons>The breeze icon theme</a> &copy; <i>The KDE Visual Design Group. (LGPL-V3)</i></li>"
    "    <li><a href=https://github.com/ColinDuquesnoy/QDarkStyleSheet>QDarkStyleSheet</a> &copy; <i>Colin Duquesnoy, Daniel Cosmo Pizetta. (MIT + CCAI-4.0)</i></li>"
    "  </ul>"
    "</body>";

AboutWindow::SimpleLabel::~SimpleLabel() noexcept {}
AboutWindow::LicenceLabel::~LicenceLabel() noexcept {}

AboutWindow::SimpleLabel::SimpleLabel(QWidget *parent, const char *text)
    : QLabel(parent)
{
    setTextFormat(Qt::RichText);
    setTextInteractionFlags(Qt::NoTextInteraction | Qt::LinksAccessibleByMouse |
                            Qt::LinksAccessibleByKeyboard);
    setText(QString::fromUtf8(text));
    setAlignment(Qt::AlignJustify | Qt::AlignTop);
    setWordWrap(true);
}

AboutWindow::LicenceLabel::LicenceLabel(QWidget *parent, const QString &url,
                                        const Qt::TextFormat format)
    : QTextEdit(parent)
{
    VIVY_LOG_CTOR() << "Creating label for licence " << VIVY_LOG_QUOTED(url);
    QFile content(url);
    if (!content.open(QIODevice::ReadOnly | QIODevice::Text))
        throw std::runtime_error("Failed to open file that should be accessible");

    setTextInteractionFlags(Qt::NoTextInteraction);
    setAlignment(Qt::AlignJustify | Qt::AlignTop);
    setAcceptRichText(true);
    setReadOnly(true);
    setAutoFormatting(QTextEdit::AutoAll);

    setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    setVerticalScrollBarPolicy(Qt::ScrollBarAsNeeded);
    setFont(vivyApp->getApplicationFont(VivyApplication::Font::Monospace));
    setFrameStyle(QFrame::NoFrame);

    switch (format) {
    case Qt::PlainText:
    case Qt::RichText: setText(QString::fromUtf8(content.readAll())); break;
    case Qt::MarkdownText: setMarkdown(QString::fromUtf8(content.readAll())); break;
    case Qt::AutoText: logError() << "Invalid text format for LicenceLabel" << format;
    }
}

// Construct the AboutWindow
AboutWindow::AboutWindow(QWidget *parent) noexcept
    : QMainWindow(parent, Qt::Dialog)
    , panels(new QTabWidget)
{
    VIVY_LOG_CTOR() << "Creating the 'Vivy - About' window";
    setWindowIcon(QIcon(VIVY_ICON_APP));
    setWindowTitle("Vivy - About");

    panels->setMovable(false);
    panels->setTabsClosable(false);
    panels->setElideMode(Qt::ElideRight);
    panels->setUsesScrollButtons(true);
    panels->setDocumentMode(false);

    panels->addTab(new SimpleLabel(panels, aboutContent), "About");
    panels->addTab(new SimpleLabel(panels, libContent), "Libraries");

    QTabWidget *licences = new QTabWidget(panels);
    licences->setTabPosition(QTabWidget::TabPosition::West);
    licences->addTab(new LicenceLabel(licences, ":licence/MIT", Qt::MarkdownText), "MIT");
    licences->addTab(new LicenceLabel(licences, ":licence/LGPL-V2.0", Qt::PlainText), "LGPL-V2.0");
    licences->addTab(new LicenceLabel(licences, ":licence/LGPL-V2.1", Qt::PlainText), "LGPL-V2.1");
    licences->addTab(new LicenceLabel(licences, ":licence/LGPL-V3", Qt::PlainText), "LGPL-V3");
    licences->addTab(new LicenceLabel(licences, ":licence/OFL-1.1", Qt::MarkdownText), "OFL-1.1");
    licences->addTab(new LicenceLabel(licences, ":licence/CCAI-4.0", Qt::MarkdownText), "CCAI-4.0");
    panels->addTab(licences, "Licences");

    setCentralWidget(panels);
    setMinimumWidth(800);
    setMinimumHeight(600);
}

void
AboutWindow::closeEvent(QCloseEvent *event) noexcept
{
    emit closed();
    event->accept();
}
