#include "PreCompiledHeaders.hh"
#include "VivyApplication.hh"
#include "UI/DocumentViews/AudioVisualizer.hh"
#include "UI/DocumentViews/AudioVisualizer/TimingParams.hh"
#include "UI/DocumentViews/AudioVisualizer/TimingBar.hh"
#include "UI/DocumentViews/AudioVisualizer/TimingView.hh"
#include "Lib/Audio.hh"

using namespace Vivy;

#define PIXMULTIPLIER 4 // Some magic AV magic stuff

AudioVisualizer::AudioVisualizer(AudioContext::StreamPtr stream, VivyDocumentView &rootView,
                                 QWidget *parent)
    : QWidget(parent)
    , AbstractVivyDocumentNeeder(rootView)
    , audioStream(stream)
{
    if (!audioStream->isDecoded()) {
        qDebug() << "Need to decode data for stream" << audioStream->getStreamIndex();
        audioStream->decodeData();
    }

    double *decodedData = audioStream->getDecodedData();
    if (decodedData == nullptr)
        throw std::logic_error("the passed stream is not decoded");

    const size_t size       = audioStream->getDecodedDataSize();
    const size_t height     = audioStream->getDecodedChunkSize();
    const size_t halfHeight = size_t(height/2);
    const size_t decalage   = audioStream->getDecodedDecalage();
    const size_t width      = (size - height) / decalage;
    uchar *pixels =
        new uchar[static_cast<size_t>(width * height / 2 + (size - height) / decalage)]();

    FFTSamplePtr chunkData(
        reinterpret_cast<FFTSample *>(av_malloc_array(2 * height, sizeof(FFTSample))),
        fftSampleDeleter);
    RDFTContextPtr ctx(av_rdft_init((static_cast<int>(log2(static_cast<int>(height)))), DFT_R2C),
                       rdftContextDeleter);

    if (!(chunkData && ctx)) {
        delete[] pixels;
        throw std::runtime_error("out of memory");
    }

    /* Compute the image data */

    for (size_t x = 0, i = 0; i < size - height; i += decalage, ++x) {
        parallel_for (size_t j = 0; j < height; j++) {
            const double curr_dat = decodedData[i + j];
            const double window_modifier =
                (1 - cos(2 * M_PI * static_cast<double>(j) / static_cast<double>(height - 1))) / 2;
            const float value =
                std::clamp(static_cast<float>(window_modifier * curr_dat), -1.0f, 1.0f);
            chunkData[j] = value;
        }

        av_rdft_calc(ctx.get(), chunkData.get());

        parallel_for (size_t j = 0; j < height / 2; j++) {
            const float im     = chunkData[j * 2];
            const float re     = chunkData[j * 2 + 1];
            const float mag    = sqrtf(im * im + re * re);
            const size_t index = static_cast<size_t>(j * static_cast<ulong>(width) + x);
            pixels[index]      = static_cast<unsigned char>(mag);
        }
    }

    QImage img = QImage(int(width), int(halfHeight), QImage::Format_RGB888);
    parallel_for (int y = 0; y < int(halfHeight); y++){
        int tempLine = y * int(width);
        parallel_for (int x = 0; x < int(width); x++){
            uint pixValue = pixels[tempLine + x];
            // TODO: See if we could make this a function provided by the user if wanted, to adapt to different spectrums
            uint pixValueAdjusted = uint(PIXMULTIPLIER * double(std::pow(pixValue/255.0, 0.3)) * 255);

            uchar r,g,b;
            b = uchar(std::min<uint>(255, pixValueAdjusted));
            if (Q_UNLIKELY(b == 255)){
                g = std::max<uchar>(0, uchar(std::min<uint>(255, pixValueAdjusted - 255)));
                r = std::max<uchar>(0, uchar(std::min<uint>(255, pixValueAdjusted - 510)));
            } else {
                g = r = 0;
            }
            img.setPixel(x, y, qRgb(r, g, b));
        }
    }
    img = img.mirrored(false, true);

    printSpectrum(img, audioStream);
}

void
AudioVisualizer::printSpectrum(QImage pixmap, AudioContext::StreamPtr stream) noexcept
{
    TimingScene *timingScene = new TimingScene(pixmap, stream, rootVivyDocumentView, this);
    TimingView *timingView   = new TimingView(timingScene, pixmap, stream, this);

    // The only that we want to take all the space is the timing scene in itself
    QGridLayout *layout = new QGridLayout;
    layout->addWidget(timingView, 1, 0);
    layout->setColumnStretch(0, 10);
    layout->addWidget(timingScene->getParams(), 1, 1);
    layout->setColumnStretch(1, 0);
    setLayout(layout);

    connect(timingScene->getParams()->getZoomSlider(), &QSlider::valueChanged,
            timingScene->getParams(), &TimingParams::setAudioWidthScale);
    // FIXME: remove for now, may freeze PC
    //connect(timingScene->getParams(), &TimingParams::paramsChanged, timingScene, &TimingScene::rebuildScene);
    connect(timingScene->getParams()->getRebuildSceneButton(), &QPushButton::released, timingScene,
            &TimingScene::rebuildScene);
    connect(&rootVivyDocumentView, &VivyDocumentView::assSubDocumentChanged, timingScene,
            &TimingScene::rebuildScene);
}

#undef PIXMULTIPLIER
