#include "TimingLine.hh"
#include "TimingParams.hh"

using namespace Vivy;

TimingLine::TimingLine(VVLib::ASSLine l[], qint64 index, QGraphicsItem *parent)
    : QGraphicsObject(parent)
    , line(l)
    , lineIndex(index)
{
}

QRectF
TimingLine::boundingRect() const
{
    const auto end   = VVLib::ASSLineGetFiniTime(line);
    const auto start = VVLib::ASSLineGetStartTime(line);
    return QRectF(tempOffset, 0, getTimingParam()->posFromMs(int(10 * (end - start))),
                  static_cast<int>(getTimingParam()->audioHeight()));
}

void
TimingLine::paint(QPainter *painter, const QStyleOptionGraphicsItem *, QWidget *)
{
    const auto end   = VVLib::ASSLineGetFiniTime(line);
    const auto start = VVLib::ASSLineGetStartTime(line);
    const QRectF rect(tempOffset, 0,
                      static_cast<int>(getTimingParam()->posFromMs(10 * (end - start))),
                      static_cast<int>(getTimingParam()->audioHeight()));
    painter->fillRect(rect, QColor(0, 255, 255, 50));
}

void
TimingLine::timingSeparatorHasChanged([[maybe_unused]] qint64 sylIndex, [[maybe_unused]] qreal x)
{
}

void
TimingLine::sepEnterPress([[maybe_unused]] qint64 sepIndex)
{
}

void
TimingLine::sepExitPress(qint64 sepIndex)
{
    if (qAbs(tempOffset) <= DBL_EPSILON) {
        prepareGeometryChange();
        if (sepIndex == 0) {
            moveBy(tempOffset, 0);

            timingSyls[sepIndex]->setPos(0, 0);
            seps[sepIndex]->silentlySetPos(0, 0);

            for (int i = 1; i < timingSyls.size(); i++) {
                timingSyls[i]->moveBy(-tempOffset, 0);
            }
            for (int i = 1; i < seps.size(); i++) {
                seps[i]->silentlyMoveBy(-tempOffset, 0);
            }

            VVLib::ASSLineSetStartTime(
                line, qint64(getTimingParam()->msFromPos(mapToScene(0, 0).x()) / 10));
        }
    }

    tempOffset = 0;
}

qreal
TimingLine::requestMove(qint64 sepIndex, qreal x)
{
    QRectF sceneRect        = mapRectFromScene(scene()->sceneRect());
    const qint64 syls_count = static_cast<qint64>(VVLib::ASSLineGetSyllabesCount(line));
    qreal mini, maxi, given = x;

    if (sepIndex <= 0) {
        prepareGeometryChange();

        mini = sceneRect.left();
        maxi = sepIndex < syls_count - 1
                   ? timingSyls[sepIndex + 1]->pos().x()
                   : getTimingParam()->posFromMs(int(VVLib::ASSLineGetDuration(line) * 10));

        given      = qBound(mini, given, maxi);
        tempOffset = given;

        qint64 dur1 = qint64(getTimingParam()->msFromPos(int(seps[1]->pos().x() - given)) / 10);
        VVLib::ASSSyllabeSetDuration(VVLib::ASSLineGetSyllabeAt(line, 0), dur1);
        timingSyls[0]->setLen(dur1);
        timingSyls[0]->setPos(given, 0);
        VVLib::ASSLineSetStartTime(
            line, static_cast<int64_t>(
                      getTimingParam()->msFromPos(mapToScene(seps[0]->pos().x(), 0).x()) / 10));
    }

    else if (sepIndex >= syls_count) {
        prepareGeometryChange();

        mini  = timingSyls[sepIndex - 1]->pos().x();
        maxi  = sceneRect.right();
        given = qBound(mini, given, maxi);

        qint64 dur2 = qint64(
            getTimingParam()->msFromPos(int(given - timingSyls[sepIndex - 1]->pos().x())) / 10);
        VVLib::ASSSyllabeSetDuration(
            VVLib::ASSLineGetSyllabeAt(line, static_cast<quint64>(sepIndex - 1)), dur2);
        timingSyls[sepIndex - 1]->setLen(dur2);
        VVLib::ASSLineSetFiniTime(line, VVLib::ASSLineGetStartTime(line) +
                                            qint64(getTimingParam()->msFromPos(int(given)) / 10));
    }

    else {
        mini  = timingSyls[sepIndex - 1]->pos().x();
        maxi  = sepIndex < syls_count - 1
                    ? timingSyls[sepIndex + 1]->pos().x()
                    : getTimingParam()->posFromMs(int(VVLib::ASSLineGetDuration(line) * 10));
        given = qBound(mini, given, maxi);

        const auto current_syl = VVLib::ASSLineGetSyllabeAt(line, static_cast<quint64>(sepIndex));
        const auto prev_syl = VVLib::ASSLineGetSyllabeAt(line, static_cast<quint64>(sepIndex - 1));

        const qint64 sumDur =
            VVLib::ASSSyllabeGetDuration(current_syl) + VVLib::ASSSyllabeGetDuration(prev_syl);
        const qint64 dur1 = qMin(
            getTimingParam()->msFromPos(int(given) - int(timingSyls[sepIndex - 1]->pos().x())) / 10,
            sumDur);
        qint64 dur2 = sumDur - dur1;

        VVLib::ASSSyllabeSetDuration(prev_syl, dur1);
        VVLib::ASSSyllabeSetDuration(current_syl, dur2);

        timingSyls[sepIndex - 1]->setLen(dur1);
        timingSyls[sepIndex]->setPos(given, 0);
        timingSyls[sepIndex]->setLen(dur2);
    }

    emit lineChanged(lineIndex);
    return given;
}

#define CONNECT_SEP(sep)                                                                           \
    connect(sep, &TimingSeparator::positionChanged, this, &TimingLine::timingSeparatorHasChanged); \
    connect(sep, &TimingSeparator::enterPress, this, &TimingLine::sepEnterPress);                  \
    connect(sep, &TimingSeparator::exitPress, this, &TimingLine::sepExitPress);
void
TimingLine::onParamsChanged()
{
    setPos(static_cast<qreal>(getTimingParam()->posFromMs(VVLib::ASSLineGetStartTime(line) * 10)),
           static_cast<qreal>(getTimingParam()->axisHeight()));
    qint64 currentTime = 0;
    qint64 endSyl      = 0;
    qint64 i;

    setZValue(Z_LINE_BACKGROUND);

    TimingSeparator *timingSeparatorStart =
        new TimingSeparator(currentTime, 0, TimingSeparator::SeparatorStyle::Start, this);
    seps.append(timingSeparatorStart);
    CONNECT_SEP(timingSeparatorStart);

    const qint64 syls_count = static_cast<qint64>(VVLib::ASSLineGetSyllabesCount(line));
    for (i = 0; i < syls_count; i += 1) {
        auto *syllabe = VVLib::ASSLineGetSyllabeAt(line, static_cast<quint64>(i));
        endSyl        = currentTime + 10 * int(VVLib::ASSSyllabeGetDuration(syllabe));

        TimingSyl *timingSyl = new TimingSyl(syllabe, currentTime, this);
        timingSyls.append(timingSyl);

        if (i != 0) {
            TimingSeparator *timingSeparator =
                new TimingSeparator(currentTime, i, TimingSeparator::SeparatorStyle::Middle, this);
            seps.append(timingSeparator);
            CONNECT_SEP(timingSeparator);
        }

        currentTime = endSyl;
    }

    TimingSeparator *timingSeparatorEnd =
        new TimingSeparator(currentTime, i, TimingSeparator::SeparatorStyle::End, this);
    seps.append(timingSeparatorEnd);
    CONNECT_SEP(timingSeparatorEnd);
}
#undef CONNECT_SEP
