#ifndef VIVY_TIMING_PARAMS_H
#define VIVY_TIMING_PARAMS_H

#ifndef __cplusplus
#error "This is a C++ header"
#endif

#include "PreCompiledHeaders.hh"
#include "Lib/Utils.hh"
#include "TimingScene.hh"

#define getTimingScene() static_cast<TimingScene *>(scene())
#define getTimingParam() getTimingScene()->getParams()

#define Z_SPECTER             (-1000)
#define Z_AXIS                (10)
#define Z_LINE_BACKGROUND     (-100)
#define Z_SEPARATOR_START_END (1000)
#define Z_SEPARATOR_MIDDLE    (100)
#define Z_LINE_SYL_TEXT       (50)
#define Z_CURSOR_BAR          (500)

namespace Vivy
{
class TimingParams final : public QWidget {
    Q_OBJECT

public:
    explicit TimingParams(QWidget * = nullptr) noexcept;

private:
    std::unique_ptr<QSlider> zoomSlider;
    std::unique_ptr<QSlider> sensibilitySlider;
    std::unique_ptr<QPushButton> rebuildSceneButton;

public:
    QSlider *getZoomSlider() const noexcept;
    QSlider *getSensibilitySlider() const noexcept;
    QPushButton *getRebuildSceneButton() const noexcept;

private:
    // Managed by TimingAxis
    qint64 m_majorTicks{ 1 }, m_minorTicks{ 1 };
    // Managed by TimingScene
    qint64 m_axisHeight{ 1 }, m_audioWidth{ 1 }, m_audioLength{ 1 }, m_audioHeight{ 1 };
    // Self-managed
    qreal m_wl{ 0 }, m_lw{ 0 };
    qreal m_audioWidthScale{ 1 };

public:
    void adjustTranslate(QRectF *, qreal, qreal) noexcept;
    void adjustFlip(QRectF *, qreal, qreal) noexcept;

    QString printMajorTicks(qint64) noexcept;
    QString printCursor(qint64) noexcept;

    inline void setTicks(qint64 t, qint64 T) noexcept
    {
        m_minorTicks = t;
        m_majorTicks = T;
    }
    inline void setAxisHeight(qint64 x) noexcept { m_axisHeight = x; }
    inline void setAudioHeight(qint64 x) noexcept { m_audioHeight = x; }
    inline void setAudioWidth(qint64 x) noexcept
    {
        m_audioWidth = x * static_cast<qint64>(m_audioWidthScale);
        updateRatios();
    }
    inline void setAudioLength(qint64 x) noexcept
    {
        m_audioLength = x;
        updateRatios();
    }

    inline qint64 axisHeight() const noexcept { return m_axisHeight; }
    inline qint64 audioHeight() const noexcept { return m_audioHeight; }
    inline qint64 audioWidth() const noexcept { return m_audioWidth; }
    inline qint64 audioLength() const noexcept { return m_audioLength; }

    inline void setAudioWidthScale(qint64 s) noexcept
    {
        m_audioWidthScale = 1.0 + 0.01 * static_cast<qreal>(s);
        setAudioWidth(m_audioWidth);
        updateRatios();
    }

    template <typename T> T posFromMs(T t) noexcept
    {
        return static_cast<T>(static_cast<qreal>(t) * m_wl);
    }

    template <typename T> T posFromMs(T t, qint64 audioWidth, qint64 audioLength) noexcept
    {
        return audioLength == 0 ? 0 : qint64(t) * audioWidth / audioLength;
    }

    template <typename T> T msFromPos(T x) noexcept
    {
        return static_cast<T>(static_cast<qreal>(x) * m_lw);
    }

    template <typename T> T msFromPos(T x, qint64 audioLength, qint64 audioWidth) noexcept
    {
        return audioWidth == 0 ? 0 : qint64(x) * audioLength / audioWidth;
    }

private:
    void updateRatios() noexcept;

signals:
    void paramsChanged();
};

}

#endif // VIVY_TIMING_PARAMS_H
