#include "PreCompiledHeaders.hh"
#include "Lib/Utils.hh"
#include "TimingScene.hh"
#include "TimingAxis.hh"
#include "TimingLine.hh"
#include "TimingParams.hh"
#include "TimingCursor.hh"

#include <QGraphicsLineItem>
#include <QGraphicsPixmapItem>
#include <QGraphicsView>
#include <QLabel>
#include <QMessageBox>
#include <QMouseEvent>
#include <QPainter>
#include <QGraphicsSceneMouseEvent>
#include <QScrollArea>
#include <QScrollBar>
#include <QVBoxLayout>

using namespace Vivy;

TimingScene::TimingScene(QImage img_, AudioContext::StreamPtr stream, VivyDocumentView &rootView,
                         QWidget *parent) noexcept
    : QGraphicsScene(parent)
    , AbstractVivyDocumentNeeder(rootView)
    , img(img_)
    , audioStream(stream)
{
    params = new TimingParams();
    // We put a common origin for the items at the bottom of the axis
    params->setAxisHeight(30); // TODO: remove magic number

    rebuildScene();
}

void
TimingScene::mousePressEvent(QGraphicsSceneMouseEvent *event) noexcept
{
    QPointF pos        = event->scenePos();
    QGraphicsItem *got = itemAt(pos, QTransform());

    if (currentLine && (got == nullptr || got == backgroundImg)) [[likely]] {
        // Handle the different cases
        if (timingMode == TimingMode::Line)
            handleMousePressEventLine(event, currentLine);
        else if (timingMode == TimingMode::Syl)
            handleMousePressEventSyl(event, currentLine);
        else if (timingMode == TimingMode::Char)
            handleMousePressEventChar(event, currentLine);
    }

    QGraphicsScene::mousePressEvent(event);
}

void
TimingScene::handleMousePressEventLine(QGraphicsSceneMouseEvent *event, VVLib::ASSLine p[]) noexcept
{
    const qint64 time = timeFromPos(static_cast<qint64>(event->scenePos().x()));
    if (const auto &btn = event->button(); btn == Qt::LeftButton) {
        VVLib::ASSLineSetStartTime(p, static_cast<int64_t>(time));
    } else if (btn == Qt::RightButton) {
        VVLib::ASSLineSetFiniTime(p, static_cast<int64_t>(time));
    }
}

void
TimingScene::handleMousePressEventSyl(QGraphicsSceneMouseEvent *, VVLib::ASSLine[]) noexcept
{
}

void
TimingScene::handleMousePressEventChar(QGraphicsSceneMouseEvent *, VVLib::ASSLine[]) noexcept
{
}

QGraphicsPixmapItem *
TimingScene::bg() noexcept
{
    return backgroundImg;
}

void
TimingScene::updateSelectedLines(const QItemSelection &selected, const QItemSelection &deselected)
{
    auto indexes = selected.indexes();
    for (int i = 0; i < indexes.size(); ++i) {
        timingLines[indexes[i].row()]->setVisible(true);
    }

    indexes = deselected.indexes();
    for (int i = 0; i < indexes.size(); ++i) {
        timingLines[indexes[i].row()]->setVisible(false);
    }
}

void
TimingScene::updateScene(QImage img_, AudioContext::StreamPtr stream_)
{
    img         = img_;
    audioStream = stream_;
    rebuildScene();
}

void
TimingScene::rebuildScene()
{
    for (auto &item : items()) {
        removeItem(item);
        delete (item);
    }
    timingLines.clear();

    const QPixmap pixmap(QPixmap::fromImage(img));
    backgroundImg = addPixmap(pixmap);
    backgroundImg->setZValue(Z_SPECTER);
    backgroundImg->setPos(0.0, static_cast<qreal>(params->axisHeight()));

    params->setAudioHeight(pixmap.height());
    params->setAudioWidth(img.width());
    params->setAudioLength(int(audioStream->getLength()));

    ax = new TimingAxis();
    addItem(ax);
    ax->onParamsChanged();
    ax->setZValue(Z_AXIS);
    ax->setPos(0.0, static_cast<qreal>(params->axisHeight()));

    cursor = new TimingCursor();
    addItem(cursor);
    cursor->onParamsChanged();
    cursor->setPos(0.0, static_cast<qreal>(params->axisHeight()));

    // Freeze the scene boundaries
    const QRectF sRect = sceneRect();
    const QRectF rect(0.0, sRect.top(), static_cast<qreal>(params->audioWidth()), sRect.height());
    setSceneRect(rect.normalized());

    if (auto assDocument = rootVivyDocument.getAssSubDocument()) {
        QVector<VVLib::ASSLine *> lines = assDocument->getLines();
        for (int i = 0; i < lines.size(); ++i) {
            if (auto line = lines.at(i)) {
                TimingLine *l = new TimingLine(line, i);
                addItem(l);
                l->onParamsChanged();
                l->setVisible(false);
                timingLines.append(l);
                if (auto assLinesModel = rootVivyDocumentView.getAssLinesModel()) {
                    connect(l, &TimingLine::lineChanged, assLinesModel, &AssLinesModel::updateLine);
                }
            }
        }
    }

    if (auto assLinesView = rootVivyDocumentView.getAssLinesView()) {
        connect(assLinesView, &AssLinesView::updateSelectedLines, this,
                &TimingScene::updateSelectedLines);
    }
}

TimingParams *
TimingScene::getParams()
{
    return params;
}

TimingAxis *
TimingScene::getAxis()
{
    return ax;
}

void
TimingScene::mouseMoveEvent(QGraphicsSceneMouseEvent *mouseEvent) noexcept
{
    int x = int(mouseEvent->scenePos().rx());
    cursor->setPos(x, 30);
    cursor->setTime(params->printCursor(params->msFromPos(x)));
    QGraphicsScene::mouseMoveEvent(mouseEvent);
}

qint64
TimingScene::timeFromPos(qint64 x) const noexcept
{
    if (const qint64 w = qint64(width()); x <= 0 || w <= 0) {
        qCritical() << "Try avoid possible divide by zero in the time from position";
        return 0;
    } else {
        return x * audioStream->getLength() / w;
    }
}

qint64
TimingScene::posFromTime(qint64 t) const noexcept
{
    if (const qint64 w = qint64(width()); t <= 0 || w <= 0) {
        qCritical() << "Try avoid possible divide by zero in the position from time";
        return 0;
    } else {
        return 10 * t * w / audioStream->getLength();
    }
}
