#include "PreCompiledHeaders.hh"
#include "UI/DocumentViews/MpvContainer.hh"
#include <QtGui/QOpenGLContext>
#include <QtCore/QMetaObject>
#include <mpv/client.h>
#include <mpv/render_gl.h>

using namespace Vivy;
using namespace std::string_literals;

static void *
get_proc_address([[maybe_unused]] void *ctx, const char *name)
{
    QOpenGLContext *glctx = QOpenGLContext::currentContext();
    if (!glctx)
        return nullptr;
    return reinterpret_cast<void *>(glctx->getProcAddress(QByteArray(name)));
}

void
MpvContainer::mpvEventWakeUpCB(void *user) noexcept
{
    QMetaObject::invokeMethod(reinterpret_cast<MpvContainer *>(user), "mpvEvent",
                              Qt::QueuedConnection);
}

MpvContainer::MpvContainer(QWidget *parent)
    : QOpenGLWidget(parent)
    , mpv(mpv_create())
{
    if (mpv == nullptr) {
        throw std::runtime_error("Failed to create the MPV context");
    } else {
        setAttribute(Qt::WA_DontCreateNativeAncestors);
        setAttribute(Qt::WA_NativeWindow);

        mpv_set_option_string(mpv, "idle", "yes");
        mpv_set_option_string(mpv, "loop-file", "inf");
        mpv_set_option_string(mpv, "no-config", "yes");
        mpv_set_option_string(mpv, "sid", "no");
        mpv_set_option_string(mpv, "input-default-bindings", "no");
        mpv_set_option_string(mpv, "input-vo-keyboard", "no");
        mpv_set_option_string(mpv, "no-ytdl", "yes");
        mpv_request_log_messages(mpv, "info");

        connect(this, &MpvContainer::mpvEvent, this, &MpvContainer::onMpvEvent,
                Qt::QueuedConnection);

        if (int rc = mpv_initialize(mpv); rc < 0) {
            printMpvError(rc);
            throw std::runtime_error("Failed to initialize the mpv context");
        }

        mpv_observe_property(mpv, 0, "pause", MPV_FORMAT_FLAG);
        mpv_observe_property(mpv, 0, "unpause", MPV_FORMAT_FLAG);
        mpv_observe_property(mpv, 0, "duration", MPV_FORMAT_DOUBLE);
        mpv_observe_property(mpv, 0, "time-pos", MPV_FORMAT_DOUBLE);
        mpv_set_wakeup_callback(mpv, &MpvContainer::mpvEventWakeUpCB, this);
    }
}

void
MpvContainer::initializeGL()
{
    mpv_opengl_init_params gl_init_params;
    memset(&gl_init_params, 0, sizeof(mpv_opengl_init_params));
    gl_init_params.get_proc_address = get_proc_address;
    mpv_render_param params[]{ { MPV_RENDER_PARAM_API_TYPE,
                                 const_cast<char *>(MPV_RENDER_API_TYPE_OPENGL) },
                               { MPV_RENDER_PARAM_OPENGL_INIT_PARAMS, &gl_init_params },
                               { MPV_RENDER_PARAM_INVALID, nullptr } };

    if (mpv_render_context_create(&mpv_gl, mpv, params) < 0) {
        throw std::runtime_error("failed to initialize mpv GL context");
    } else {
        mpv_render_context_set_update_callback(mpv_gl, MpvContainer::onUpdate,
                                               reinterpret_cast<void *>(this));
    }
}

void
MpvContainer::paintGL() noexcept
{
    mpv_opengl_fbo mpfbo{ static_cast<int>(defaultFramebufferObject()), width(), height(), 0 };
    int flip_y{ 1 };

    mpv_render_param params[] = { { MPV_RENDER_PARAM_OPENGL_FBO, &mpfbo },
                                  { MPV_RENDER_PARAM_FLIP_Y, &flip_y },
                                  { MPV_RENDER_PARAM_INVALID, nullptr } };
    // See render_gl.h on what OpenGL environment mpv expects, and
    // other API details.
    mpv_render_context_render(mpv_gl, params);
}

void
MpvContainer::onUpdate(void *ctx) noexcept
{
    QMetaObject::invokeMethod(reinterpret_cast<MpvContainer *>(ctx), "maybeUpdate");
}

void
MpvContainer::registerMpvTimeCallback(std::function<void(double)> callback) noexcept
{
    if (mpvTimeCallback)
        logWarning() << "Override a previous mpv callback!";
    mpvTimeCallback = callback;
}

void
MpvContainer::registerMpvDurationCallback(std::function<void(double)> callback) noexcept
{
    if (mpvDurationCallback)
        logWarning() << "Override a previous mpv callback!";
    mpvDurationCallback = callback;
}

MpvContainer::~MpvContainer() noexcept
{
    makeCurrent();
    logDebug() << "Closing the MPV context";
    if (mpv_gl) {
        mpv_render_context_free(mpv_gl);
    }
    if (mpv) {
        asyncCommand(AsyncCmdType::None, { "quit", nullptr });
        mpv_wait_async_requests(mpv);
        registerMpvTimeCallback(nullptr);
        registerMpvDurationCallback(nullptr);
        mpv_destroy(mpv);
        mpv = nullptr; // Stop all other callbacks here
    }
}

void
MpvContainer::handleMpvEvent(const mpv_event *const event) noexcept
{
    // Declare here variables that can be used in the switch-case statements
    double time;
    QString msgText;
    union {
        mpv_event_log_message *msg;
        mpv_event_property *prop;
    };

    auto checkProp = [](const mpv_event_property *const prop_, const std::string_view str,
                        const int format) noexcept -> bool {
        return (prop_->name == str) && (prop_->format == format);
    };

    switch (event->event_id) {
    case MPV_EVENT_SHUTDOWN: logWarning() << "Ignore shutdown event for now..."; break;

    case MPV_EVENT_LOG_MESSAGE:
        msg     = reinterpret_cast<mpv_event_log_message *>(event->data);
        msgText = QString::fromUtf8(msg->text);
        msgText.replace('\n', "");
        logDebug() << "MPV - MSG [" << msg->prefix << "] " << msg->level << ": " << msgText;
        break;

    case MPV_EVENT_PROPERTY_CHANGE:
        prop = reinterpret_cast<mpv_event_property *>(event->data);
        if (checkProp(prop, "time-pos", MPV_FORMAT_DOUBLE) && mpvTimeCallback) {
            time = *reinterpret_cast<double *>(prop->data);
            mpvTimeCallback(time);
        }

        else if (checkProp(prop, "duration", MPV_FORMAT_DOUBLE) && mpvDurationCallback) {
            time = *reinterpret_cast<double *>(prop->data);
            mpvDurationCallback(time);
        }

        else if (checkProp(prop, "pause", MPV_FORMAT_FLAG)) {
            isPlaybackPaused = *reinterpret_cast<bool *>(prop->data);
            emit mpvPlaybackToggled(!isPlaybackPaused);
            logDebug() << "MPV -> set to " << (isPlaybackPaused ? "pause" : "play");
        }

        else if (checkProp(prop, "unpause", MPV_FORMAT_FLAG)) {
            isPlaybackPaused = !*reinterpret_cast<bool *>(prop->data);
            emit mpvPlaybackToggled(!isPlaybackPaused);
            logDebug() << "MPV -> set to " << (isPlaybackPaused ? "pause" : "play");
        }

        break;

    case MPV_EVENT_COMMAND_REPLY:
        handleMpvEventCommandReply(static_cast<AsyncCmdType>(event->reply_userdata));
        break;

    // Explicitly ignored
    case MPV_EVENT_START_FILE:
    case MPV_EVENT_END_FILE:
    case MPV_EVENT_VIDEO_RECONFIG:
    case MPV_EVENT_NONE:
    case MPV_EVENT_GET_PROPERTY_REPLY:
    case MPV_EVENT_SET_PROPERTY_REPLY:
    case MPV_EVENT_FILE_LOADED:
    case MPV_EVENT_CLIENT_MESSAGE:
    case MPV_EVENT_AUDIO_RECONFIG:
    case MPV_EVENT_SEEK:
    case MPV_EVENT_PLAYBACK_RESTART:
    case MPV_EVENT_QUEUE_OVERFLOW:
    case MPV_EVENT_HOOK: break;
    }
}

int
MpvContainer::getAssSid() const noexcept
{
    bool conversionOk  = false;
    const char *result = mpv_get_property_string(mpv, "sid");
    const int ret      = QString::fromUtf8(result).toInt(&conversionOk);
    return (result == nullptr || !conversionOk) ? -1 : ret;
}

void
MpvContainer::handleMpvEventCommandReply(const AsyncCmdType type) noexcept
{
    switch (type) {
    case AsyncCmdType::None: break;

    case AsyncCmdType::LoadAssFile:
    case AsyncCmdType::ReloadAss:
        sid = getAssSid();
        logDebug() << "Load ASS file with id:" << sid;
        break;

    case AsyncCmdType::UnloadAss:
        sid = getAssSid();
        logDebug() << "Unload Ass, rc = " << sid;
        if (sid != -1)
            unloadAssFile();
        break;

    case AsyncCmdType::LoadFile:
        sid = getAssSid();
        logDebug() << "MPV - CMD: File loaded by mpv, sid =" << sid;
        isPlaybackPaused = false;
        mpvPause();
        unloadAssFile();
        break;

    case AsyncCmdType::SeekTime: break;
    case AsyncCmdType::TogglePlayback: break;
    }
}

void
MpvContainer::onMpvEvent() noexcept
{
    while (mpv) {
        const mpv_event *const event = mpv_wait_event(mpv, 0);
        if (event == nullptr || event->event_id == MPV_EVENT_NONE)
            break;
        handleMpvEvent(event);
    }

    if (mpv == nullptr) {
        logDebug() << "MPV was closed while in the event loop!";
    }
}

// Make Qt invoke mpv_render_context_render() to draw a new/updated video frame.
void
MpvContainer::maybeUpdate() noexcept
{
    // If the Qt window is not visible, Qt's update() will just skip rendering.
    // This confuses mpv's render API, and may lead to small occasional
    // freezes due to video rendering timing out.
    // Handle this by manually redrawing.
    // Note: Qt doesn't seem to provide a way to query whether update() will
    //       be skipped, and the following code still fails when e.g. switching
    //       to a different workspace with a reparenting window manager.
    if (window()->isMinimized()) {
        makeCurrent();
        paintGL();
        context()->swapBuffers(context()->surface());
        doneCurrent();
    } else {
        update();
    }
}

void
MpvContainer::loadFile(const QString &filename) noexcept
{
    if (filename.size() == 0)
        return;

    qDebug() << "Loading file" << filename;
    if (previousLoadedFile != filename)
        previousLoadedFile = filename;
    const QByteArray cFileName = filename.toUtf8();
    asyncCommand(AsyncCmdType::LoadFile, { "loadfile", cFileName.data(), nullptr });
}

void
MpvContainer::loadAssFile(const QString &ass) noexcept
{
    if (ass.isEmpty())
        return;

    const QByteArray cFileName = ass.toUtf8();
    asyncCommand(AsyncCmdType::LoadAssFile, { "sub-add", cFileName.data(), nullptr });
}

void
MpvContainer::reloadAssFile() noexcept
{
    asyncCommand(AsyncCmdType::ReloadAss, { "sub-reload", nullptr });
}

void
MpvContainer::printMpvError(int rc) const noexcept
{
    if (rc == MPV_ERROR_SUCCESS)
        return;

    logError() << "MPV error: " << mpv_error_string(rc);
}

void
MpvContainer::mpvPlay() noexcept
{
    if (isPlaybackPaused)
        mpvTogglePlayback();
}

void
MpvContainer::mpvPause() noexcept
{
    if (!isPlaybackPaused)
        mpvTogglePlayback();
}

void
MpvContainer::mpvTogglePlayback() noexcept
{
    logDebug() << "MPV: Toggling the playback";
    asyncCommand(AsyncCmdType::TogglePlayback, { "cycle", "pause", "up", nullptr });
}

void
MpvContainer::asyncCommand(const AsyncCmdType cmd,
                           std::initializer_list<const char *> args) noexcept
{
    // NOTE: const_cast here, we have faith in MPV to not change the value of
    // the temporary pointers here. Should be OK anyway because the `args` init
    // list is a temporary and should be discarded afer the method call.
    printMpvError(mpv_command_async(mpv, cmd, const_cast<const char **>(std::data(args))));
}

void
MpvContainer::unloadAssFile() noexcept
{
    asyncCommand(AsyncCmdType::UnloadAss, { "sub-remove", nullptr });
}

void
MpvContainer::seekInFile(const chrono::seconds time) noexcept
{
    QByteArray seconds = QString::number(time.count()).toUtf8();
    asyncCommand(AsyncCmdType::SeekTime, { "seek", seconds.data(), "absolute", nullptr });
}
